# -*- coding: utf8 -*-
# flake8: NOQA
import sys

PY3 = sys.version_info[0] == 3
PYMINOR = sys.version_info[1]
PYPATCH = sys.version_info[2]

_identity = lambda x: x

if PY3 and PYMINOR >= 7:
    breakpoint = breakpoint
else:
    import pdb

    breakpoint = pdb.set_trace


console_encoding = sys.__stdout__.encoding

implements_to_string = _identity


def console_to_str(s):
    """From pypa/pip project, pip.backwardwardcompat. License MIT."""
    try:
        return s.decode(console_encoding)
    except UnicodeDecodeError:
        return s.decode('utf_8')



from tmuxp.plugin import TmuxpPlugin


class PluginAfterWindowFinished(TmuxpPlugin):
    def __init__(self):
        self.message = '[+] This is the Tmuxp Test Plugin'

    def after_window_finished(self, window):
        if window.name == 'editor':
            window.rename_window('plugin_test_awf')
        elif window.name == 'awf_mw_test':
            window.rename_window('plugin_test_awf_mw')
        elif window.name == 'awf_mw_test_2':
            window.rename_window('plugin_test_awf_mw_2')
        elif window.name == 'mp_test_owc':
            window.rename_window('mp_test_awf')
        else:
            pass



"""Exceptions for tmuxp.

tmuxp.exc
~~~~~~~~~

"""
from ._compat import implements_to_string


class TmuxpException(Exception):

    """Base Exception for Tmuxp Errors."""

    pass


class ConfigError(TmuxpException):

    """Error parsing tmuxp configuration dict."""

    pass


class EmptyConfigException(ConfigError):

    """Configuration is empty."""

    pass


class TmuxpPluginException(TmuxpException):

    """Base Exception for Tmuxp Errors."""

    pass


class BeforeLoadScriptNotExists(OSError):
    def __init__(self, *args, **kwargs):
        super(BeforeLoadScriptNotExists, self).__init__(*args, **kwargs)

        self.strerror = "before_script file '%s' doesn't exist." % self.strerror


@implements_to_string
class BeforeLoadScriptError(Exception):

    """Exception replacing :py:class:`subprocess.CalledProcessError` for
    :meth:`tmuxp.util.run_before_script`.
    """

    def __init__(self, returncode, cmd, output=None):
        self.returncode = returncode
        self.cmd = cmd
        self.output = output
        self.message = (
            'before_script failed with returncode {returncode}.\n'
            'command: {cmd}\n'
            'Error output:\n'
            '{output}'
        ).format(returncode=self.returncode, cmd=self.cmd, output=self.output)

    def __str__(self):
        return self.message



import os

config_unexpanded = {  # shell_command_before is string in some areas
    'session_name': 'sampleconfig',
    'start_directory': '/',
    'windows': [
        {
            'window_name': 'editor',
            'start_directory': '~',
            'shell_command_before': 'source .venv/bin/activate',
            'panes': [
                {'shell_command': ['vim']},
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': ['cowsay "hey"'],
                },
            ],
            'layout': 'main-verticle',
        },
        {
            'shell_command_before': 'rbenv local 2.0.0-p0',
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}, {}],
        },
        {
            'window_name': 'shufu',
            'panes': [
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': ['htop'],
                }
            ],
        },
        {'options': {'automatic-rename': True}, 'panes': [{'shell_command': ['htop']}]},
        {'panes': ['top']},
    ],
}

config_expanded = {  # shell_command_before is string in some areas
    'session_name': 'sampleconfig',
    'start_directory': '/',
    'windows': [
        {
            'window_name': 'editor',
            'start_directory': os.path.expanduser('~'),
            'shell_command_before': ['source .venv/bin/activate'],
            'panes': [
                {'shell_command': ['vim']},
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': ['cowsay "hey"'],
                },
            ],
            'layout': 'main-verticle',
        },
        {
            'shell_command_before': ['rbenv local 2.0.0-p0'],
            'window_name': 'logging',
            'panes': [
                {'shell_command': ['tail -F /var/log/syslog']},
                {'shell_command': []},
            ],
        },
        {
            'window_name': 'shufu',
            'panes': [
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': ['htop'],
                }
            ],
        },
        {'options': {'automatic-rename': True}, 'panes': [{'shell_command': ['htop']}]},
        {'panes': [{'shell_command': ['top']}]},
    ],
}

config_after = {  # shell_command_before is string in some areas
    'session_name': 'sampleconfig',
    'start_directory': '/',
    'windows': [
        {
            'window_name': 'editor',
            'start_directory': os.path.expanduser('~'),
            'shell_command_before': ['source .venv/bin/activate'],
            'panes': [
                {'shell_command': ['source .venv/bin/activate', 'vim']},
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': [
                        'source .venv/bin/activate',
                        'rbenv local 2.0.0-p0',
                        'cowsay "hey"',
                    ],
                },
            ],
            'layout': 'main-verticle',
        },
        {
            'shell_command_before': ['rbenv local 2.0.0-p0'],
            'start_directory': '/',
            'window_name': 'logging',
            'panes': [
                {'shell_command': ['rbenv local 2.0.0-p0', 'tail -F /var/log/syslog']},
                {'shell_command': ['rbenv local 2.0.0-p0']},
            ],
        },
        {
            'start_directory': '/',
            'window_name': 'shufu',
            'panes': [
                {
                    'shell_command_before': ['rbenv local 2.0.0-p0'],
                    'shell_command': ['rbenv local 2.0.0-p0', 'htop'],
                }
            ],
        },
        {
            'start_directory': '/',
            'options': {'automatic-rename': True},
            'panes': [{'shell_command': ['htop']}],
        },
        {'start_directory': '/', 'panes': [{'shell_command': ['top']}]},
    ],
}



from tmuxp.plugin import TmuxpPlugin


class PluginFailVersion(TmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'tmuxp-plugin-fail-version',
            'tmuxp_max_version': '0.0.0',
        }
        TmuxpPlugin.__init__(self, **config)



import os

from .._util import loadfixture

unexpanded_yaml = loadfixture('config/expand2-unexpanded.yaml')
expanded_yaml = loadfixture('config/expand2-expanded.yaml').format(
    HOME=os.path.expanduser('~')
)



"""Tests for tmuxp.

tmuxp.tests
~~~~~~~~~~~

"""
import os

current_dir = os.path.abspath(os.path.dirname(__file__))
example_dir = os.path.abspath(os.path.join(current_dir, '..', 'examples'))
fixtures_dir = os.path.realpath(os.path.join(current_dir, 'fixtures'))



from tmuxp.plugin import TmuxpPlugin


class PluginOnWindowCreate(TmuxpPlugin):
    def __init__(self):
        self.message = '[+] This is the Tmuxp Test Plugin'

    def on_window_create(self, window):
        if window.name == 'editor':
            window.rename_window('plugin_test_owc')
        elif window.name == 'owc_mw_test':
            window.rename_window('plugin_test_owc_mw')
        elif window.name == 'owc_mw_test_2':
            window.rename_window('plugin_test_owc_mw_2')
        elif window.name == 'mp_test':
            window.rename_window('mp_test_owc')
        else:
            pass



from .._util import loadfixture

tmuxinator_yaml = loadfixture('config_tmuxinator/test3.yaml')

tmuxinator_dict = {
    'name': 'sample',
    'root': '~/test',
    'socket_name': 'foo',
    'tmux_options': '-f ~/.tmux.mac.conf',
    'pre': 'sudo /etc/rc.d/mysqld start',
    'pre_window': 'rbenv shell 2.0.0-p247',
    'windows': [
        {
            'editor': {
                'pre': [
                    'echo "I get run in each pane, ' 'before each pane command!"',
                    None,
                ],
                'layout': 'main-vertical',
                'root': '~/test/editor',
                'panes': ['vim', None, 'top'],
            }
        },
        {'shell': ['git pull', 'git merge']},
        {
            'guard': {
                'layout': 'tiled',
                'pre': [
                    'echo "I get run in each pane."',
                    'echo "Before each pane command!"',
                ],
                'panes': [None, None, None],
            }
        },
        {'database': 'bundle exec rails db'},
        {'server': 'bundle exec rails s'},
        {'logs': 'tail -f log/development.log'},
        {'console': 'bundle exec rails c'},
        {'capistrano': None},
        {'server': 'ssh user@example.com'},
    ],
}

expected = {
    'session_name': 'sample',
    'socket_name': 'foo',
    'start_directory': '~/test',
    'config': '~/.tmux.mac.conf',
    'shell_command': 'sudo /etc/rc.d/mysqld start',
    'shell_command_before': ['rbenv shell 2.0.0-p247'],
    'windows': [
        {
            'window_name': 'editor',
            'shell_command_before': [
                'echo "I get run in each pane, before each pane command!"',
                None,
            ],
            'layout': 'main-vertical',
            'start_directory': '~/test/editor',
            'panes': ['vim', None, 'top'],
        },
        {'window_name': 'shell', 'panes': ['git pull', 'git merge']},
        {
            'window_name': 'guard',
            'layout': 'tiled',
            'shell_command_before': [
                'echo "I get run in each pane."',
                'echo "Before each pane command!"',
            ],
            'panes': [None, None, None],
        },
        {'window_name': 'database', 'panes': ['bundle exec rails db']},
        {'window_name': 'server', 'panes': ['bundle exec rails s']},
        {'window_name': 'logs', 'panes': ['tail -f log/development.log']},
        {'window_name': 'console', 'panes': ['bundle exec rails c']},
        {'window_name': 'capistrano', 'panes': [None]},
        {'window_name': 'server', 'panes': ['ssh user@example.com']},
    ],
}



#!/usr/bin/env python
"""Log utilities for tmuxp.

tmuxp.log
~~~~~~~~~

"""
import logging
import time

from colorama import Fore, Style

LEVEL_COLORS = {
    'DEBUG': Fore.BLUE,  # Blue
    'INFO': Fore.GREEN,  # Green
    'WARNING': Fore.YELLOW,
    'ERROR': Fore.RED,
    'CRITICAL': Fore.RED,
}

LOG_LEVELS = {
    'CRITICAL': 50,
    'ERROR': 40,
    'WARNING': 30,
    'INFO': 20,
    'DEBUG': 10,
    'NOTSET': 0,
}


def set_style(
    message, stylized, style_before=None, style_after=None, prefix='', suffix=''
):
    if stylized:
        return prefix + style_before + message + style_after + suffix

    return prefix + message + suffix


def default_log_template(self, record, stylized=False):
    """
    Return the prefix for the log message. Template for Formatter.

    Parameters
    ----------
    :py:class:`logging.LogRecord` :
        object. this is passed in from inside the
        :py:meth:`logging.Formatter.format` record.

    Returns
    -------
    str
        template for logger message
    """

    reset = Style.RESET_ALL
    levelname = set_style(
        '(%(levelname)s)',
        stylized,
        style_before=(LEVEL_COLORS.get(record.levelname) + Style.BRIGHT),
        style_after=Style.RESET_ALL,
        suffix=' ',
    )
    asctime = set_style(
        '%(asctime)s',
        stylized,
        style_before=(Fore.BLACK + Style.DIM + Style.BRIGHT),
        style_after=(Fore.RESET + Style.RESET_ALL),
        prefix='[',
        suffix=']',
    )
    name = set_style(
        '%(name)s',
        stylized,
        style_before=(Fore.WHITE + Style.DIM + Style.BRIGHT),
        style_after=(Fore.RESET + Style.RESET_ALL),
        prefix=' ',
        suffix=' ',
    )

    if stylized:
        return reset + levelname + asctime + name + reset

    return levelname + asctime + name


class LogFormatter(logging.Formatter):
    template = default_log_template

    def __init__(self, color=True, *args, **kwargs):
        logging.Formatter.__init__(self, *args, **kwargs)

    def format(self, record):
        try:
            record.message = record.getMessage()
        except Exception as e:
            record.message = "Bad message (%r): %r" % (e, record.__dict__)

        date_format = '%H:%m:%S'
        record.asctime = time.strftime(date_format, self.converter(record.created))

        prefix = self.template(record) % record.__dict__

        parts = prefix.split(record.message)
        formatted = prefix + " " + record.message
        return formatted.replace("\n", "\n" + parts[0] + " ")


def debug_log_template(self, record):
    """
    Return the prefix for the log message. Template for Formatter.

    Parameters
    ----------
    record :  :py:class:`logging.LogRecord`
        This is passed in from inside the :py:meth:`logging.Formatter.format`
        record.

    Returns
    -------
    str
        Log template.
    """

    reset = Style.RESET_ALL
    levelname = (
        LEVEL_COLORS.get(record.levelname)
        + Style.BRIGHT
        + '(%(levelname)1.1s)'
        + Style.RESET_ALL
        + ' '
    )
    asctime = (
        '['
        + Fore.BLACK
        + Style.DIM
        + Style.BRIGHT
        + '%(asctime)s'
        + Fore.RESET
        + Style.RESET_ALL
        + ']'
    )
    name = (
        ' '
        + Fore.WHITE
        + Style.DIM
        + Style.BRIGHT
        + '%(name)s'
        + Fore.RESET
        + Style.RESET_ALL
        + ' '
    )
    module_funcName = Fore.GREEN + Style.BRIGHT + '%(module)s.%(funcName)s()'
    lineno = (
        Fore.BLACK
        + Style.DIM
        + Style.BRIGHT
        + ':'
        + Style.RESET_ALL
        + Fore.CYAN
        + '%(lineno)d'
    )

    tpl = reset + levelname + asctime + name + module_funcName + lineno + reset

    return tpl


class DebugLogFormatter(LogFormatter):
    """Provides greater technical details than standard log Formatter."""

    template = debug_log_template



from .._util import loadfixture

teamocil_yaml = loadfixture('config_teamocil/test3.yaml')

teamocil_dict = {
    'windows': [
        {
            'name': 'my-first-window',
            'root': '~/Projects/foo-www',
            'layout': 'even-vertical',
            'filters': {
                'before': 'rbenv local 2.0.0-p0',
                'after': 'echo \'I am done initializing this pane.\'',
            },
            'panes': [
                {'cmd': 'git status'},
                {'cmd': 'bundle exec rails server --port 40', 'focus': True},
                {'cmd': ['sudo service memcached start', 'sudo service mongodb start']},
            ],
        }
    ]
}

expected = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'my-first-window',
            'layout': 'even-vertical',
            'start_directory': "~/Projects/foo-www",
            'shell_command_before': 'rbenv local 2.0.0-p0',
            'shell_command_after': ('echo ' '\'I am done initializing this pane.\''),
            'panes': [
                {'shell_command': 'git status'},
                {'shell_command': 'bundle exec rails server --port 40', 'focus': True},
                {
                    'shell_command': [
                        'sudo service memcached start',
                        'sudo service mongodb start',
                    ]
                },
            ],
        }
    ],
}



import logging
import os

import pytest

from libtmux import exc
from libtmux.server import Server
from libtmux.test import TEST_SESSION_PREFIX, get_test_session_name, namer

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def monkeypatch_plugin_test_packages(monkeypatch):
    paths = [
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bwb/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bs/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_r/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_owc/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_awf/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_fail/",
    ]
    for path in paths:
        monkeypatch.syspath_prepend(os.path.abspath(os.path.relpath(path)))


@pytest.fixture(scope='function')
def socket_name(request):
    return 'tmuxp_test%s' % next(namer)


@pytest.fixture(scope='function')
def server(request, socket_name):
    t = Server()
    t.socket_name = socket_name

    def fin():
        t.kill_server()

    request.addfinalizer(fin)

    return t


@pytest.fixture(scope='function')
def session(server):
    session_name = 'tmuxp'

    if not server.has_session(session_name):
        server.cmd(
            '-f',
            '/dev/null',  # use a blank config to reduce side effects
            'new-session',
            '-d',  # detached
            '-s',
            session_name,
            '/bin/sh',  # use /bin/sh as a shell to reduce side effects
            # normally, it'd be -c, but new-session is special
        )

    # find current sessions prefixed with tmuxp
    old_test_sessions = [
        s.get('session_name')
        for s in server._sessions
        if s.get('session_name').startswith(TEST_SESSION_PREFIX)
    ]

    TEST_SESSION_NAME = get_test_session_name(server=server)

    try:
        session = server.new_session(session_name=TEST_SESSION_NAME)
    except exc.LibTmuxException as e:
        raise e

    """
    Make sure that tmuxp can :ref:`test_builder_visually` and switches to
    the newly created session for that testcase.
    """
    try:
        server.switch_client(session.get('session_id'))
        pass
    except exc.LibTmuxException:
        # server.attach_session(session.get('session_id'))
        pass

    for old_test_session in old_test_sessions:
        logger.debug('Old test test session %s found. Killing it.' % old_test_session)
        server.kill_session(old_test_session)
    assert TEST_SESSION_NAME == session.get('session_name')
    assert TEST_SESSION_NAME != 'tmuxp'

    return session


@pytest.fixture()
def tmpdir(tmpdir_factory):
    fn = tmpdir_factory.mktemp('tmuxp')
    return fn



from tmuxp import shell


def test_detect_best_shell():
    result = shell.detect_best_shell()
    assert isinstance(result, str)


def test_shell_detect():
    assert isinstance(shell.has_bpython(), bool)
    assert isinstance(shell.has_ipython(), bool)
    assert isinstance(shell.has_ptpython(), bool)
    assert isinstance(shell.has_ptipython(), bool)



from .test_plugin_helpers import MyTestTmuxpPlugin


class LibtmuxVersionFailMinPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'libtmux-min-version-fail',
            'libtmux_min_version': '0.8.3',
            'libtmux_version': '0.7.0',
        }
        MyTestTmuxpPlugin.__init__(self, config)


class LibtmuxVersionFailMaxPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'libtmux-max-version-fail',
            'libtmux_max_version': '3.0',
            'libtmux_version': '3.5',
        }
        MyTestTmuxpPlugin.__init__(self, config)


class LibtmuxVersionFailIncompatiblePlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'libtmux-incompatible-version-fail',
            'libtmux_version_incompatible': ['0.7.1'],
            'libtmux_version': '0.7.1',
        }
        MyTestTmuxpPlugin.__init__(self, config)



"""Test for tmuxp configuration import, inlining, expanding and export."""
import os

import pytest

import kaptan

from tmuxp import config, exc

from . import example_dir
from .fixtures import config as fixtures

TMUXP_DIR = os.path.join(os.path.dirname(__file__), '.tmuxp')


def load_yaml(yaml):
    return kaptan.Kaptan(handler='yaml').import_config(yaml).get()


def load_config(_file):
    return kaptan.Kaptan().import_config(_file).get()


def test_export_json(tmpdir):
    json_config_file = tmpdir.join('config.json')

    configparser = kaptan.Kaptan()
    configparser.import_config(fixtures.sampleconfig.sampleconfigdict)

    json_config_data = configparser.export('json', indent=2)

    json_config_file.write(json_config_data)

    new_config = kaptan.Kaptan()
    new_config_data = new_config.import_config(str(json_config_file)).get()
    assert fixtures.sampleconfig.sampleconfigdict == new_config_data


def test_export_yaml(tmpdir):
    yaml_config_file = tmpdir.join('config.yaml')

    configparser = kaptan.Kaptan()
    sampleconfig = config.inline(fixtures.sampleconfig.sampleconfigdict)
    configparser.import_config(sampleconfig)

    yaml_config_data = configparser.export('yaml', indent=2, default_flow_style=False)

    yaml_config_file.write(yaml_config_data)

    new_config_data = load_config(str(yaml_config_file))
    assert fixtures.sampleconfig.sampleconfigdict == new_config_data


def test_scan_config(tmpdir):
    configs = []

    garbage_file = tmpdir.join('config.psd')
    garbage_file.write('wat')

    for r, d, f in os.walk(str(tmpdir)):
        for filela in (x for x in f if x.endswith(('.json', '.ini', 'yaml'))):
            configs.append(str(tmpdir.join(filela)))

    files = 0
    if tmpdir.join('config.json').check():
        files += 1
        assert str(tmpdir.join('config.json')) in configs

    if tmpdir.join('config.yaml').check():
        files += 1
        assert str(tmpdir.join('config.yaml')) in configs

    if tmpdir.join('config.ini').check():
        files += 1
        assert str(tmpdir.join('config.ini')) in configs

    assert len(configs) == files


def test_config_expand1():
    """Expand shell commands from string to list."""
    test_config = config.expand(fixtures.expand1.before_config)
    assert test_config == fixtures.expand1.after_config


def test_config_expand2():
    """Expand shell commands from string to list."""

    unexpanded_dict = load_yaml(fixtures.expand2.unexpanded_yaml)

    expanded_dict = load_yaml(fixtures.expand2.expanded_yaml)

    assert config.expand(unexpanded_dict) == expanded_dict


"""Tests for :meth:`config.inline()`."""

ibefore_config = {  # inline config
    'session_name': 'sampleconfig',
    'start_directory': '~',
    'windows': [
        {
            'shell_command': ['top'],
            'window_name': 'editor',
            'panes': [{'shell_command': ['vim']}, {'shell_command': ['cowsay "hey"']}],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}],
        },
        {'options': {'automatic-rename': True}, 'panes': [{'shell_command': ['htop']}]},
    ],
}

iafter_config = {
    'session_name': 'sampleconfig',
    'start_directory': '~',
    'windows': [
        {
            'shell_command': 'top',
            'window_name': 'editor',
            'panes': ['vim', 'cowsay "hey"'],
            'layout': 'main-verticle',
        },
        {'window_name': 'logging', 'panes': ['tail -F /var/log/syslog']},
        {'options': {'automatic-rename': True}, 'panes': ['htop']},
    ],
}


def test_inline_config():
    """:meth:`config.inline()` shell commands list to string."""

    test_config = config.inline(ibefore_config)
    assert test_config == iafter_config


"""Test config inheritance for the nested 'start_command'."""

inheritance_config_before = {
    'session_name': 'sampleconfig',
    'start_directory': '/',
    'windows': [
        {
            'window_name': 'editor',
            'start_directory': '~',
            'panes': [{'shell_command': ['vim']}, {'shell_command': ['cowsay "hey"']}],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}],
        },
        {'window_name': 'shufu', 'panes': [{'shell_command': ['htop']}]},
        {'options': {'automatic-rename': True}, 'panes': [{'shell_command': ['htop']}]},
    ],
}

inheritance_config_after = {
    'session_name': 'sampleconfig',
    'start_directory': '/',
    'windows': [
        {
            'window_name': 'editor',
            'start_directory': '~',
            'panes': [{'shell_command': ['vim']}, {'shell_command': ['cowsay "hey"']}],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}],
        },
        {'window_name': 'shufu', 'panes': [{'shell_command': ['htop']}]},
        {'options': {'automatic-rename': True}, 'panes': [{'shell_command': ['htop']}]},
    ],
}


def test_inheritance_config():
    config = inheritance_config_before

    # TODO: Look at verifying window_start_directory
    # if 'start_directory' in config:
    #     session_start_directory = config['start_directory']
    # else:
    #     session_start_directory = None

    # for windowconfitem in config['windows']:
    #     window_start_directory = None
    #
    #     if 'start_directory' in windowconfitem:
    #         window_start_directory = windowconfitem['start_directory']
    #     elif session_start_directory:
    #         window_start_directory = session_start_directory
    #
    #     for paneconfitem in windowconfitem['panes']:
    #         if 'start_directory' in paneconfitem:
    #             pane_start_directory = paneconfitem['start_directory']
    #         elif window_start_directory:
    #             paneconfitem['start_directory'] = window_start_directory
    #         elif session_start_directory:
    #             paneconfitem['start_directory'] = session_start_directory

    assert config == inheritance_config_after


def test_shell_command_before():
    """Config inheritance for the nested 'start_command'."""
    test_config = fixtures.shell_command_before.config_unexpanded
    test_config = config.expand(test_config)

    assert test_config == fixtures.shell_command_before.config_expanded

    test_config = config.trickle(test_config)
    assert test_config == fixtures.shell_command_before.config_after


def test_in_session_scope():
    sconfig = load_yaml(fixtures.shell_command_before_session.before)

    config.validate_schema(sconfig)

    assert config.expand(sconfig) == sconfig
    assert config.expand(config.trickle(sconfig)) == load_yaml(
        fixtures.shell_command_before_session.expected
    )


def test_trickle_relative_start_directory():
    test_config = config.trickle(fixtures.trickle.before)
    assert test_config == fixtures.trickle.expected


def test_trickle_window_with_no_pane_config():
    test_yaml = """
    session_name: test_session
    windows:
    - window_name: test_1
      panes:
      - shell_command:
        - ls -l
    - window_name: test_no_panes
    """
    sconfig = load_yaml(test_yaml)
    config.validate_schema(sconfig)

    assert config.expand(config.trickle(sconfig))['windows'][1]['panes'][0] == {
        'shell_command': []
    }


def test_expands_blank_panes():
    """Expand blank config into full form.

    Handle ``NoneType`` and 'blank'::

    # nothing, None, 'blank'
    'panes': [
        None,
        'blank'
    ]

    # should be blank
    'panes': [
        'shell_command': []
    ]

    Blank strings::

        panes: [
            ''
        ]

        # should output to:
        panes:
            'shell_command': ['']

    """

    yaml_config_file = os.path.join(example_dir, 'blank-panes.yaml')
    test_config = load_config(yaml_config_file)
    assert config.expand(test_config) == fixtures.expand_blank.expected


def test_no_session_name():
    yaml_config = """
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - window_name: logging
      automatic-rename: true
      panes:
      - shell_command:
      - htop
    """

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches(r'requires "session_name"')


def test_no_windows():
    yaml_config = """
    session_name: test session
    """

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.match(r'list of "windows"')


def test_no_window_name():
    yaml_config = """
    session_name: test session
    windows:
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - automatic-rename: true
      panes:
      - shell_command:
      - htop
    """

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches('missing "window_name"')


def test_replaces_env_variables(monkeypatch):
    env_key = "TESTHEY92"
    env_val = "HEYO1"
    yaml_config = """
    start_directory: {TEST_VAR}/test
    shell_command_before: {TEST_VAR}/test2
    before_script: {TEST_VAR}/test3
    session_name: hi - {TEST_VAR}
    options:
        default-command: {TEST_VAR}/lol
    global_options:
        default-shell: {TEST_VAR}/moo
    windows:
    - window_name: editor
      panes:
      - shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - window_name: logging @ {TEST_VAR}
      automatic-rename: true
      panes:
      - shell_command:
      - htop
    """.format(
        TEST_VAR="${%s}" % env_key
    )

    sconfig = load_yaml(yaml_config)

    monkeypatch.setenv(str(env_key), str(env_val))
    sconfig = config.expand(sconfig)
    assert "%s/test" % env_val == sconfig['start_directory']
    assert "%s/test2" % env_val in sconfig['shell_command_before']
    assert "%s/test3" % env_val == sconfig['before_script']
    assert "hi - %s" % env_val == sconfig['session_name']
    assert "%s/moo" % env_val == sconfig['global_options']['default-shell']
    assert "%s/lol" % env_val == sconfig['options']['default-command']
    assert "logging @ %s" % env_val == sconfig['windows'][1]['window_name']


def test_plugins():
    yaml_config = """
    session_name: test session
    plugins: tmuxp-plugin-one.plugin.TestPluginOne
    windows:
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    """

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches('only supports list type')



from .._util import loadfixture

teamocil_yaml = loadfixture('config_teamocil/test2.yaml')
teamocil_dict = {
    'windows': [
        {
            'name': 'sample-four-panes',
            'root': '~/Code/sample/www',
            'layout': 'tiled',
            'panes': [{'cmd': 'pwd'}, {'cmd': 'pwd'}, {'cmd': 'pwd'}, {'cmd': 'pwd'}],
        }
    ]
}

expected = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'sample-four-panes',
            'layout': 'tiled',
            'start_directory': '~/Code/sample/www',
            'panes': [
                {'shell_command': 'pwd'},
                {'shell_command': 'pwd'},
                {'shell_command': 'pwd'},
                {'shell_command': 'pwd'},
            ],
        }
    ],
}



from .._util import loadfixture

tmuxinator_yaml = loadfixture('config_tmuxinator/test1.yaml')
tmuxinator_dict = {
    'windows': [
        {'editor': {'layout': 'main-vertical', 'panes': ['vim', 'guard']}},
        {'server': 'bundle exec rails s'},
        {'logs': 'tail -f logs/development.log'},
    ]
}

expected = {
    'session_name': None,
    'windows': [
        {'window_name': 'editor', 'layout': 'main-vertical', 'panes': ['vim', 'guard']},
        {'window_name': 'server', 'panes': ['bundle exec rails s']},
        {'window_name': 'logs', 'panes': ['tail -f logs/development.log']},
    ],
}



sampleconfigdict = {
    'session_name': 'sampleconfig',
    'start_directory': '~',
    'windows': [
        {
            'window_name': 'editor',
            'panes': [
                {'start_directory': '~', 'shell_command': ['vim']},
                {'shell_command': ['cowsay "hey"']},
            ],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [
                {
                    'shell_command': ['tail -F /var/log/syslog'],
                    'start_directory': '/var/log',
                }
            ],
        },
        {'options': {'automatic_rename': True}, 'panes': [{'shell_command': ['htop']}]},
    ],
}



import os

before_config = {
    'session_name': 'sampleconfig',
    'start_directory': '~',
    'windows': [
        {
            'window_name': 'editor',
            'panes': [
                {'shell_command': ['vim', 'top']},
                {'shell_command': ['vim']},
                {'shell_command': 'cowsay "hey"'},
            ],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}],
        },
        {
            'start_directory': '/var/log',
            'options': {'automatic-rename': True},
            'panes': [{'shell_command': 'htop'}, 'vim'],
        },
        {'start_directory': './', 'panes': ['pwd']},
        {'start_directory': './asdf/', 'panes': ['pwd']},
        {'start_directory': '../', 'panes': ['pwd']},
        {'panes': ['top']},
    ],
}

after_config = {
    'session_name': 'sampleconfig',
    'start_directory': os.path.expanduser('~'),
    'windows': [
        {
            'window_name': 'editor',
            'panes': [
                {'shell_command': ['vim', 'top']},
                {'shell_command': ['vim']},
                {'shell_command': ['cowsay "hey"']},
            ],
            'layout': 'main-verticle',
        },
        {
            'window_name': 'logging',
            'panes': [{'shell_command': ['tail -F /var/log/syslog']}],
        },
        {
            'start_directory': '/var/log',
            'options': {'automatic-rename': True},
            'panes': [{'shell_command': ['htop']}, {'shell_command': ['vim']}],
        },
        {
            'start_directory': os.path.normpath(
                os.path.join(os.path.expanduser('~'), './')
            ),
            'panes': [{'shell_command': ['pwd']}],
        },
        {
            'start_directory': os.path.normpath(
                os.path.join(os.path.expanduser('~'), './asdf')
            ),
            'panes': [{'shell_command': ['pwd']}],
        },
        {
            'start_directory': os.path.normpath(
                os.path.join(os.path.expanduser('~'), '../')
            ),
            'panes': [{'shell_command': ['pwd']}],
        },
        {'panes': [{'shell_command': ['top']}]},
    ],
}



from tmuxp.plugin import TmuxpPlugin


class MyTestTmuxpPlugin(TmuxpPlugin):
    def __init__(self, config):
        tmux_version = config.pop('tmux_version', None)
        libtmux_version = config.pop('libtmux_version', None)
        tmuxp_version = config.pop('tmuxp_version', None)

        TmuxpPlugin.__init__(self, **config)

        # WARNING! This should not be done in anything but a test
        if tmux_version:
            self.version_constraints['tmux']['version'] = tmux_version
        if libtmux_version:
            self.version_constraints['libtmux']['version'] = libtmux_version
        if tmuxp_version:
            self.version_constraints['tmuxp']['version'] = tmuxp_version

        self._version_check()



"""Configuration parsing and export for tmuxp.

tmuxp.config
~~~~~~~~~~~~

"""
import copy
import logging
import os

from . import exc

logger = logging.getLogger(__name__)


def validate_schema(sconf):
    """
    Return True if config schema is correct.

    Parameters
    ----------
    sconf : dict
        session configuration

    Returns
    -------
    bool
    """

    # verify session_name
    if 'session_name' not in sconf:
        raise exc.ConfigError('config requires "session_name"')

    if 'windows' not in sconf:
        raise exc.ConfigError('config requires list of "windows"')

    for window in sconf['windows']:
        if 'window_name' not in window:
            raise exc.ConfigError('config window is missing "window_name"')

    if 'plugins' in sconf:
        if not isinstance(sconf['plugins'], list):
            raise exc.ConfigError('"plugins" only supports list type')

    return True


def is_config_file(filename, extensions=['.yml', '.yaml', '.json']):
    """
    Return True if file has a valid config file type.

    Parameters
    ----------
    filename : str
        filename to check (e.g. ``mysession.json``).
    extensions : str or list
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    bool
    """
    extensions = [extensions] if isinstance(extensions, str) else extensions
    return any(filename.endswith(e) for e in extensions)


def in_dir(
    config_dir=os.path.expanduser('~/.tmuxp'), extensions=['.yml', '.yaml', '.json']
):
    """
    Return a list of configs in ``config_dir``.

    Parameters
    ----------
    config_dir : str
        directory to search
    extensions : list
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    list
    """
    configs = []

    for filename in os.listdir(config_dir):
        if is_config_file(filename, extensions) and not filename.startswith('.'):
            configs.append(filename)

    return configs


def in_cwd():
    """
    Return list of configs in current working directory.

    If filename is ``.tmuxp.py``, ``.tmuxp.json``, ``.tmuxp.yaml``.

    Returns
    -------
    list
        configs in current working directory
    """
    configs = []

    for filename in os.listdir(os.getcwd()):
        if filename.startswith('.tmuxp') and is_config_file(filename):
            configs.append(filename)

    return configs


def expandshell(_path):
    """
    Return expanded path based on user's ``$HOME`` and ``env``.

    :py:func:`os.path.expanduser` and :py:func:`os.path.expandvars`

    Parameters
    ----------
    path : str
        path to expand

    Returns
    -------
    str
        path with shell variables expanded
    """
    return os.path.expandvars(os.path.expanduser(_path))


def inline(sconf):
    """
    Return config in inline form, opposite of :meth:`config.expand`.

    Parameters
    ----------
    sconf : dict

    Returns
    -------
    dict
        configuration with optional inlined configs.
    """

    if (
        'shell_command' in sconf
        and isinstance(sconf['shell_command'], list)
        and len(sconf['shell_command']) == 1
    ):
        sconf['shell_command'] = sconf['shell_command'][0]

        if len(sconf.keys()) == int(1):
            sconf = sconf['shell_command']
    if (
        'shell_command_before' in sconf
        and isinstance(sconf['shell_command_before'], list)
        and len(sconf['shell_command_before']) == 1
    ):
        sconf['shell_command_before'] = sconf['shell_command_before'][0]

    # recurse into window and pane config items
    if 'windows' in sconf:
        sconf['windows'] = [inline(window) for window in sconf['windows']]
    if 'panes' in sconf:
        sconf['panes'] = [inline(pane) for pane in sconf['panes']]

    return sconf


def expand(sconf, cwd=None, parent=None):
    """Return config with shorthand and inline properties expanded.

    This is necessary to keep the code in the :class:`WorkspaceBuilder` clean
    and also allow for neat, short-hand configurations.

    As a simple example, internally, tmuxp expects that config options
    like ``shell_command`` are a list (array)::

        'shell_command': ['htop']

    tmuxp configs allow for it to be simply a string::

        'shell_command': 'htop'

    Kaptan will load JSON/YAML files into python dicts for you.

    Parameters
    ----------
    sconf : dict
        the configuration for the session
    cwd : str
        directory to expand relative paths against. should be the dir of the
        config directory.
    parent : str
        (used on recursive entries) start_directory of parent window or session
        object.

    Returns
    -------
    dict
    """

    # Note: cli.py will expand configs relative to project's config directory
    # for the first cwd argument.
    if not cwd:
        cwd = os.getcwd()

    if 'session_name' in sconf:
        sconf['session_name'] = expandshell(sconf['session_name'])
    if 'window_name' in sconf:
        sconf['window_name'] = expandshell(sconf['window_name'])
    if 'environment' in sconf:
        for key in sconf['environment']:
            val = sconf['environment'][key]
            val = expandshell(val)
            if any(val.startswith(a) for a in ['.', './']):
                val = os.path.normpath(os.path.join(cwd, val))
            sconf['environment'][key] = val
    if 'global_options' in sconf:
        for key in sconf['global_options']:
            val = sconf['global_options'][key]
            if isinstance(val, str):
                val = expandshell(val)
                if any(val.startswith(a) for a in ['.', './']):
                    val = os.path.normpath(os.path.join(cwd, val))
            sconf['global_options'][key] = val
    if 'options' in sconf:
        for key in sconf['options']:
            val = sconf['options'][key]
            if isinstance(val, str):
                val = expandshell(val)
                if any(val.startswith(a) for a in ['.', './']):
                    val = os.path.normpath(os.path.join(cwd, val))
            sconf['options'][key] = val

    # Any config section, session, window, pane that can contain the
    # 'shell_command' value
    if 'start_directory' in sconf:
        sconf['start_directory'] = expandshell(sconf['start_directory'])
        start_path = sconf['start_directory']
        if any(start_path.startswith(a) for a in ['.', './']):
            # if window has a session, or pane has a window with a
            # start_directory of . or ./, make sure the start_directory can be
            # relative to the parent.
            #
            # This is for the case where you may be loading a config from
            # outside your shell current directory.
            if parent:
                cwd = parent['start_directory']
            start_path = os.path.normpath(os.path.join(cwd, start_path))
            sconf['start_directory'] = start_path

    if 'before_script' in sconf:
        sconf['before_script'] = expandshell(sconf['before_script'])
        if any(sconf['before_script'].startswith(a) for a in ['.', './']):
            sconf['before_script'] = os.path.normpath(
                os.path.join(cwd, sconf['before_script'])
            )

    if 'shell_command' in sconf and isinstance(sconf['shell_command'], str):
        sconf['shell_command'] = [sconf['shell_command']]

    if 'shell_command_before' in sconf and isinstance(
        sconf['shell_command_before'], str
    ):
        sconf['shell_command_before'] = [sconf['shell_command_before']]

    if 'shell_command_before' in sconf and isinstance(
        sconf['shell_command_before'], list
    ):
        sconf['shell_command_before'] = [
            expandshell(scmd) for scmd in sconf['shell_command_before']
        ]

    # recurse into window and pane config items
    if 'windows' in sconf:
        sconf['windows'] = [expand(window, parent=sconf) for window in sconf['windows']]
    elif 'panes' in sconf:

        for pconf in sconf['panes']:
            p_index = sconf['panes'].index(pconf)
            p = copy.deepcopy(pconf)
            pconf = sconf['panes'][p_index] = {}

            if isinstance(p, str):
                p = {'shell_command': [p]}
            elif not p:
                p = {'shell_command': []}

            assert isinstance(p, dict)
            if 'shell_command' in p:
                cmd = p['shell_command']

                if isinstance(p['shell_command'], str):
                    cmd = [cmd]

                if not cmd or any(a == cmd for a in [None, 'blank', 'pane']):
                    cmd = []

                if isinstance(cmd, list) and len(cmd) == int(1):
                    if any(a in cmd for a in [None, 'blank', 'pane']):
                        cmd = []

                p['shell_command'] = cmd
            else:
                p['shell_command'] = []

            pconf.update(p)
        sconf['panes'] = [expand(pane, parent=sconf) for pane in sconf['panes']]

    return sconf


def trickle(sconf):
    """Return a dict with "trickled down" / inherited config values.

    This will only work if config has been expanded to full form with
    :meth:`config.expand`.

    tmuxp allows certain commands to be default at the session, window
    level. shell_command_before trickles down and prepends the
    ``shell_command`` for the pane.

    Parameters
    ----------
    sconf : dict
        the session configuration.

    Returns
    -------
    dict
    """

    # prepends a pane's ``shell_command`` list with the window and sessions'
    # ``shell_command_before``.

    if 'start_directory' in sconf:
        session_start_directory = sconf['start_directory']
    else:
        session_start_directory = None

    if 'suppress_history' in sconf:
        suppress_history = sconf['suppress_history']
    else:
        suppress_history = None

    for windowconfig in sconf['windows']:

        # Prepend start_directory to relative window commands
        if session_start_directory:
            if 'start_directory' not in windowconfig:
                windowconfig['start_directory'] = session_start_directory
            else:
                if not any(
                    windowconfig['start_directory'].startswith(a) for a in ['~', '/']
                ):
                    window_start_path = os.path.join(
                        session_start_directory, windowconfig['start_directory']
                    )
                    windowconfig['start_directory'] = window_start_path

        # We only need to trickle to the window, workspace builder checks wconf
        if suppress_history is not None:
            if 'suppress_history' not in windowconfig:
                windowconfig['suppress_history'] = suppress_history

        # If panes were NOT specified for a window, assume that a single pane
        # with no shell commands is desired
        if 'panes' not in windowconfig:
            windowconfig['panes'] = [{'shell_command': []}]

        for paneconfig in windowconfig['panes']:
            commands_before = []

            # Prepend shell_command_before to commands
            if 'shell_command_before' in sconf:
                commands_before.extend(sconf['shell_command_before'])
            if 'shell_command_before' in windowconfig:
                commands_before.extend(windowconfig['shell_command_before'])
            if 'shell_command_before' in paneconfig:
                commands_before.extend(paneconfig['shell_command_before'])

            if 'shell_command' in paneconfig:
                commands_before.extend(paneconfig['shell_command'])

            p_index = windowconfig['panes'].index(paneconfig)
            windowconfig['panes'][p_index]['shell_command'] = commands_before
            # paneconfig['shell_command'] = commands_before

    return sconf


def import_tmuxinator(sconf):
    """Return tmuxp config from a `tmuxinator`_ yaml config.

    .. _tmuxinator: https://github.com/aziz/tmuxinator

    Parameters
    ----------
    sconf : dict
        python dict for session configuration.

    Returns
    -------
    dict
    """

    tmuxp_config = {}

    if 'project_name' in sconf:
        tmuxp_config['session_name'] = sconf.pop('project_name')
    elif 'name' in sconf:
        tmuxp_config['session_name'] = sconf.pop('name')
    else:
        tmuxp_config['session_name'] = None

    if 'project_root' in sconf:
        tmuxp_config['start_directory'] = sconf.pop('project_root')
    elif 'root' in sconf:
        tmuxp_config['start_directory'] = sconf.pop('root')

    if 'cli_args' in sconf:
        tmuxp_config['config'] = sconf['cli_args']

        if '-f' in tmuxp_config['config']:
            tmuxp_config['config'] = tmuxp_config['config'].replace('-f', '').strip()
    elif 'tmux_options' in sconf:
        tmuxp_config['config'] = sconf['tmux_options']

        if '-f' in tmuxp_config['config']:
            tmuxp_config['config'] = tmuxp_config['config'].replace('-f', '').strip()

    if 'socket_name' in sconf:
        tmuxp_config['socket_name'] = sconf['socket_name']

    tmuxp_config['windows'] = []

    if 'tabs' in sconf:
        sconf['windows'] = sconf.pop('tabs')

    if 'pre' in sconf and 'pre_window' in sconf:
        tmuxp_config['shell_command'] = sconf['pre']

        if isinstance(sconf['pre'], str):
            tmuxp_config['shell_command_before'] = [sconf['pre_window']]
        else:
            tmuxp_config['shell_command_before'] = sconf['pre_window']
    elif 'pre' in sconf:
        if isinstance(sconf['pre'], str):
            tmuxp_config['shell_command_before'] = [sconf['pre']]
        else:
            tmuxp_config['shell_command_before'] = sconf['pre']

    if 'rbenv' in sconf:
        if 'shell_command_before' not in tmuxp_config:
            tmuxp_config['shell_command_before'] = []
        tmuxp_config['shell_command_before'].append('rbenv shell %s' % sconf['rbenv'])

    for w in sconf['windows']:
        for k, v in w.items():

            windowdict = {'window_name': k}

            if isinstance(v, str) or v is None:
                windowdict['panes'] = [v]
                tmuxp_config['windows'].append(windowdict)
                continue
            elif isinstance(v, list):
                windowdict['panes'] = v
                tmuxp_config['windows'].append(windowdict)
                continue

            if 'pre' in v:
                windowdict['shell_command_before'] = v['pre']
            if 'panes' in v:
                windowdict['panes'] = v['panes']
            if 'root' in v:
                windowdict['start_directory'] = v['root']

            if 'layout' in v:
                windowdict['layout'] = v['layout']
            tmuxp_config['windows'].append(windowdict)
    return tmuxp_config


def import_teamocil(sconf):
    """Return tmuxp config from a `teamocil`_ yaml config.

    .. _teamocil: https://github.com/remiprev/teamocil

    Parameters
    ----------
    sconf : dict
        python dict for session configuration

    Notes
    -----

    Todos:

    - change  'root' to a cd or start_directory
    - width in pane -> main-pain-width
    - with_env_var
    - clear
    - cmd_separator
    """

    tmuxp_config = {}

    if 'session' in sconf:
        sconf = sconf['session']

    if 'name' in sconf:
        tmuxp_config['session_name'] = sconf['name']
    else:
        tmuxp_config['session_name'] = None

    if 'root' in sconf:
        tmuxp_config['start_directory'] = sconf.pop('root')

    tmuxp_config['windows'] = []

    for w in sconf['windows']:

        windowdict = {'window_name': w['name']}

        if 'clear' in w:
            windowdict['clear'] = w['clear']

        if 'filters' in w:
            if 'before' in w['filters']:
                for b in w['filters']['before']:
                    windowdict['shell_command_before'] = w['filters']['before']
            if 'after' in w['filters']:
                for b in w['filters']['after']:
                    windowdict['shell_command_after'] = w['filters']['after']

        if 'root' in w:
            windowdict['start_directory'] = w.pop('root')

        if 'splits' in w:
            w['panes'] = w.pop('splits')

        if 'panes' in w:
            for p in w['panes']:
                if 'cmd' in p:
                    p['shell_command'] = p.pop('cmd')
                if 'width' in p:
                    # todo support for height/width
                    p.pop('width')
            windowdict['panes'] = w['panes']

        if 'layout' in w:
            windowdict['layout'] = w['layout']
        tmuxp_config['windows'].append(windowdict)

    return tmuxp_config



"""Utility and helper methods for tmuxp.

tmuxp.shell
~~~~~~~~~~~

"""
import logging
import os

logger = logging.getLogger(__name__)


def has_ipython():
    try:
        from IPython import start_ipython  # NOQA F841
    except ImportError:
        try:
            from IPython.Shell import IPShell  # NOQA F841
        except ImportError:
            return False

    return True


def has_ptpython():
    try:
        from ptpython.repl import embed, run_config  # NOQA F841
    except ImportError:
        try:
            from prompt_toolkit.contrib.repl import embed, run_config  # NOQA F841
        except ImportError:
            return False

    return True


def has_ptipython():
    try:
        from ptpython.ipython import embed  # NOQA F841
        from ptpython.repl import run_config  # NOQA F841
    except ImportError:
        try:
            from prompt_toolkit.contrib.ipython import embed  # NOQA F841
            from prompt_toolkit.contrib.repl import run_config  # NOQA F841
        except ImportError:
            return False

    return True


def has_bpython():
    try:
        from bpython import embed  # NOQA F841
    except ImportError:
        return False
    return True


def detect_best_shell():
    if has_ptipython():
        return 'ptipython'
    elif has_ptpython():
        return 'ptpython'
    elif has_ipython():
        return 'ipython'
    elif has_bpython():
        return 'bpython'
    return 'code'


def get_bpython(options, extra_args=None):
    if extra_args is None:
        extra_args = {}

    from bpython import embed  # NOQA F841

    def launch_bpython():
        imported_objects = get_launch_args(**options)
        kwargs = {}
        if extra_args:
            kwargs['args'] = extra_args
        embed(imported_objects, **kwargs)

    return launch_bpython


def get_ipython_arguments():
    ipython_args = 'IPYTHON_ARGUMENTS'
    return os.environ.get(ipython_args, '').split()


def get_ipython(options, **extra_args):
    try:
        from IPython import start_ipython

        def launch_ipython():
            imported_objects = get_launch_args(**options)
            ipython_arguments = extra_args or get_ipython_arguments()
            start_ipython(argv=ipython_arguments, user_ns=imported_objects)

        return launch_ipython
    except ImportError:
        # IPython < 0.11
        # Explicitly pass an empty list as arguments, because otherwise
        # IPython would use sys.argv from this script.
        # Notebook not supported for IPython < 0.11.
        from IPython.Shell import IPShell

        def launch_ipython():
            imported_objects = get_launch_args(**options)
            shell = IPShell(argv=[], user_ns=imported_objects)
            shell.mainloop()

        return launch_ipython


def get_ptpython(options, vi_mode=False):
    try:
        from ptpython.repl import embed, run_config
    except ImportError:
        from prompt_toolkit.contrib.repl import embed, run_config

    def launch_ptpython():
        imported_objects = get_launch_args(**options)
        history_filename = os.path.expanduser('~/.ptpython_history')
        embed(
            globals=imported_objects,
            history_filename=history_filename,
            vi_mode=vi_mode,
            configure=run_config,
        )

    return launch_ptpython


def get_ptipython(options, vi_mode=False):
    """Based on django-extensions

    Run renamed to launch, get_imported_objects renamed to get_launch_args
    """
    try:
        from ptpython.ipython import embed
        from ptpython.repl import run_config
    except ImportError:
        # prompt_toolkit < v0.27
        from prompt_toolkit.contrib.ipython import embed
        from prompt_toolkit.contrib.repl import run_config

    def launch_ptipython():
        imported_objects = get_launch_args(**options)
        history_filename = os.path.expanduser('~/.ptpython_history')
        embed(
            user_ns=imported_objects,
            history_filename=history_filename,
            vi_mode=vi_mode,
            configure=run_config,
        )

    return launch_ptipython


def get_launch_args(**kwargs):
    import libtmux

    return {
        'libtmux': libtmux,
        'Server': libtmux.Server,
        'Session': libtmux.Session,
        'Window': libtmux.Window,
        'Pane': libtmux.Pane,
        'server': kwargs.get('server'),
        'session': kwargs.get('session'),
        'window': kwargs.get('window'),
        'pane': kwargs.get('pane'),
    }


def get_code(use_pythonrc, imported_objects):
    import code

    try:
        # Try activating rlcompleter, because it's handy.
        import readline
    except ImportError:
        pass
    else:
        # We don't have to wrap the following import in a 'try', because
        # we already know 'readline' was imported successfully.
        import rlcompleter

        readline.set_completer(rlcompleter.Completer(imported_objects).complete)
        # Enable tab completion on systems using libedit (e.g. macOS).
        # These lines are copied from Lib/site.py on Python 3.4.
        readline_doc = getattr(readline, '__doc__', '')
        if readline_doc is not None and 'libedit' in readline_doc:
            readline.parse_and_bind("bind ^I rl_complete")
        else:
            readline.parse_and_bind("tab:complete")

    # We want to honor both $PYTHONSTARTUP and .pythonrc.py, so follow system
    # conventions and get $PYTHONSTARTUP first then .pythonrc.py.
    if use_pythonrc:
        for pythonrc in set(
            [os.environ.get("PYTHONSTARTUP"), os.path.expanduser('~/.pythonrc.py')]
        ):
            if not pythonrc:
                continue
            if not os.path.isfile(pythonrc):
                continue
            with open(pythonrc) as handle:
                pythonrc_code = handle.read()
            # Match the behavior of the cpython shell where an error in
            # PYTHONSTARTUP prints an exception and continues.
            exec(compile(pythonrc_code, pythonrc, 'exec'), imported_objects)

    def launch_code():
        code.interact(local=imported_objects)

    return launch_code


def launch(shell='best', use_pythonrc=False, use_vi_mode=False, **kwargs):
    # Also allowing passing shell='code' to force using code.interact
    imported_objects = get_launch_args(**kwargs)

    if shell == 'best':
        shell = detect_best_shell()

    if shell == 'ptipython':
        launch = get_ptipython(options=kwargs, vi_mode=use_vi_mode)
    elif shell == 'ptpython':
        launch = get_ptpython(options=kwargs, vi_mode=use_vi_mode)
    elif shell == 'ipython':
        launch = get_ipython(options=kwargs)
    elif shell == 'bpython':
        launch = get_bpython(options=kwargs)
    else:
        launch = get_code(use_pythonrc=use_pythonrc, imported_objects=imported_objects)

    launch()



from .._util import loadfixture

teamocil_yaml = loadfixture('config_teamocil/test1.yaml')
teamocil_conf = {
    'windows': [
        {
            'name': 'sample-two-panes',
            'root': '~/Code/sample/www',
            'layout': 'even-horizontal',
            'panes': [{'cmd': ['pwd', 'ls -la']}, {'cmd': 'rails server --port 3000'}],
        }
    ]
}

expected = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'sample-two-panes',
            'layout': 'even-horizontal',
            'start_directory': '~/Code/sample/www',
            'panes': [
                {'shell_command': ['pwd', 'ls -la']},
                {'shell_command': 'rails server --port 3000'},
            ],
        }
    ],
}



from .._util import loadfixture

teamocil_yaml = loadfixture('config_teamocil/layouts.yaml')

teamocil_dict = {
    'two-windows': {
        'windows': [
            {
                'name': 'foo',
                'clear': True,
                'root': '/foo',
                'layout': 'tiled',
                'panes': [{'cmd': "echo 'foo'"}, {'cmd': "echo 'foo again'"}],
            },
            {
                'name': 'bar',
                'root': '/bar',
                'splits': [
                    {
                        'cmd': ["echo 'bar'", "echo 'bar in an array'"],
                        'target': 'bottom-right',
                    },
                    {'cmd': "echo 'bar again'", 'focus': True, 'width': 50},
                ],
            },
        ]
    },
    'two-windows-with-filters': {
        'windows': [
            {
                'name': 'foo',
                'root': '/foo',
                'filters': {
                    'before': ['echo first before filter', 'echo second before filter'],
                    'after': ['echo first after filter', 'echo second after filter'],
                },
                'panes': [
                    {'cmd': "echo 'foo'"},
                    {'cmd': "echo 'foo again'", 'width': 50},
                ],
            }
        ]
    },
    'two-windows-with-custom-command-options': {
        'windows': [
            {
                'name': 'foo',
                'cmd_separator': '\n',
                'with_env_var': False,
                'clear': True,
                'root': '/foo',
                'layout': 'tiled',
                'panes': [{'cmd': "echo 'foo'"}, {'cmd': "echo 'foo again'"}],
            },
            {
                'name': 'bar',
                'cmd_separator': ' && ',
                'with_env_var': True,
                'root': '/bar',
                'splits': [
                    {
                        'cmd': ["echo 'bar'", "echo 'bar in an array'"],
                        'target': 'bottom-right',
                    },
                    {'cmd': "echo 'bar again'", 'focus': True, 'width': 50},
                ],
            },
        ]
    },
    'three-windows-within-a-session': {
        'session': {
            'name': 'my awesome session',
            'windows': [
                {'name': 'first window', 'panes': [{'cmd': "echo 'foo'"}]},
                {'name': 'second window', 'panes': [{'cmd': "echo 'foo'"}]},
                {'name': 'third window', 'panes': [{'cmd': "echo 'foo'"}]},
            ],
        }
    },
}


two_windows = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'foo',
            'start_directory': '/foo',
            'clear': True,
            'layout': 'tiled',
            'panes': [
                {'shell_command': "echo 'foo'"},
                {'shell_command': "echo 'foo again'"},
            ],
        },
        {
            'window_name': 'bar',
            'start_directory': '/bar',
            'panes': [
                {
                    'shell_command': ["echo 'bar'", "echo 'bar in an array'"],
                    'target': 'bottom-right',
                },
                {'shell_command': "echo 'bar again'", 'focus': True},
            ],
        },
    ],
}

two_windows_with_filters = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'foo',
            'start_directory': '/foo',
            'shell_command_before': [
                'echo first before filter',
                'echo second before filter',
            ],
            'shell_command_after': [
                'echo first after filter',
                'echo second after filter',
            ],
            'panes': [
                {'shell_command': "echo 'foo'"},
                {'shell_command': "echo 'foo again'"},
            ],
        }
    ],
}

two_windows_with_custom_command_options = {
    'session_name': None,
    'windows': [
        {
            'window_name': 'foo',
            'start_directory': '/foo',
            'clear': True,
            'layout': 'tiled',
            'panes': [
                {'shell_command': "echo 'foo'"},
                {'shell_command': "echo 'foo again'"},
            ],
        },
        {
            'window_name': 'bar',
            'start_directory': '/bar',
            'panes': [
                {
                    'shell_command': ["echo 'bar'", "echo 'bar in an array'"],
                    'target': 'bottom-right',
                },
                {'shell_command': "echo 'bar again'", 'focus': True},
            ],
        },
    ],
}

three_windows_within_a_session = {
    'session_name': 'my awesome session',
    'windows': [
        {'window_name': 'first window', 'panes': [{'shell_command': "echo 'foo'"}]},
        {'window_name': 'second window', 'panes': [{'shell_command': "echo 'foo'"}]},
        {'window_name': 'third window', 'panes': [{'shell_command': "echo 'foo'"}]},
    ],
}



from .test_plugin_helpers import MyTestTmuxpPlugin


class TmuxVersionFailMinPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'tmux-min-version-fail',
            'tmux_min_version': '1.8',
            'tmux_version': '1.7',
        }
        MyTestTmuxpPlugin.__init__(self, config)


class TmuxVersionFailMaxPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'tmux-max-version-fail',
            'tmux_max_version': '3.0',
            'tmux_version': '3.5',
        }
        MyTestTmuxpPlugin.__init__(self, config)


class TmuxVersionFailIncompatiblePlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'tmux-incompatible-version-fail',
            'tmux_version_incompatible': ['2.3'],
            'tmux_version': '2.3',
        }

        MyTestTmuxpPlugin.__init__(self, config)



from .test_plugin_helpers import MyTestTmuxpPlugin


class AllVersionPassPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            'plugin_name': 'tmuxp-plugin-my-tmuxp-plugin',
            'tmux_min_version': '1.8',
            'tmux_max_version': '100.0',
            'tmux_version_incompatible': ['2.3'],
            'libtmux_min_version': '0.8.3',
            'libtmux_max_version': '100.0',
            'libtmux_version_incompatible': ['0.7.1'],
            'tmuxp_min_version': '1.7.0',
            'tmuxp_max_version': '100.0.0',
            'tmuxp_version_incompatible': ['1.5.6'],
            'tmux_version': '3.0',
            'tmuxp_version': '1.7.0',
        }
        MyTestTmuxpPlugin.__init__(self, config)



import os


def curjoin(_file):  # return filepath relative to __file__ (this file)
    return os.path.join(os.path.dirname(__file__), _file)


def loadfixture(_file):  # return fixture data, relative to __file__
    return open(curjoin(_file)).read()



from tmuxp.plugin import TmuxpPlugin


class PluginBeforeWorkspaceBuilder(TmuxpPlugin):
    def __init__(self):
        self.message = '[+] This is the Tmuxp Test Plugin'

    def before_workspace_builder(self, session):
        session.rename_session('plugin_test_bwb')



expected = {
    'session_name': 'Blank pane test',
    'windows': [
        {
            'window_name': 'Blank pane test',
            'panes': [
                {'shell_command': []},
                {'shell_command': []},
                {'shell_command': []},
            ],
        },
        {
            'window_name': 'More blank panes',
            'panes': [
                {'shell_command': []},
                {'shell_command': []},
                {'shell_command': []},
            ],
        },
        {
            'window_name': 'Empty string (return)',
            'panes': [
                {'shell_command': ['']},
                {'shell_command': ['']},
                {'shell_command': ['']},
            ],
        },
        {
            'window_name': 'Blank with options',
            'panes': [
                {'shell_command': [], 'focus': True},
                {'shell_command': [], 'start_directory': '/tmp'},
            ],
        },
    ],
}



"""Tests for utility functions in tmux."""
import os

import pytest

from tmuxp import exc
from tmuxp.exc import BeforeLoadScriptError, BeforeLoadScriptNotExists
from tmuxp.util import run_before_script

from . import fixtures_dir


def test_raise_BeforeLoadScriptNotExists_if_not_exists():
    script_file = os.path.join(fixtures_dir, 'script_noexists.sh')

    with pytest.raises(BeforeLoadScriptNotExists):
        run_before_script(script_file)

    with pytest.raises(OSError):
        run_before_script(script_file)


def test_raise_BeforeLoadScriptError_if_retcode():
    script_file = os.path.join(fixtures_dir, 'script_failed.sh')

    with pytest.raises(BeforeLoadScriptError):
        run_before_script(script_file)


def test_return_stdout_if_ok(capsys):
    script_file = os.path.join(fixtures_dir, 'script_complete.sh')

    run_before_script(script_file)
    out, err = capsys.readouterr()
    assert 'hello' in out


def test_beforeload_returncode():
    script_file = os.path.join(fixtures_dir, 'script_failed.sh')

    with pytest.raises(exc.BeforeLoadScriptError) as excinfo:
        run_before_script(script_file)
        assert excinfo.match(r'113')


def test_beforeload_returns_stderr_messages():
    script_file = os.path.join(fixtures_dir, 'script_failed.sh')

    with pytest.raises(exc.BeforeLoadScriptError) as excinfo:
        run_before_script(script_file)
        assert excinfo.match(r'failed with returncode')



"""Test for tmuxp command line interface."""
import json
import os

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import MagicMock

import pytest

import click
import kaptan
from click.testing import CliRunner

import libtmux
from libtmux.common import has_lt_version
from libtmux.exc import LibTmuxException
from tmuxp import cli, config, exc
from tmuxp.cli import (
    _load_append_windows_to_current_session,
    _load_attached,
    _reattach,
    command_debug_info,
    command_ls,
    get_config_dir,
    is_pure_name,
    load_plugins,
    load_workspace,
    scan_config,
)
from tmuxp.workspacebuilder import WorkspaceBuilder

from .fixtures._util import curjoin, loadfixture


def test_creates_config_dir_not_exists(tmpdir):
    """cli.startup() creates config dir if not exists."""

    cli.startup(str(tmpdir))
    assert os.path.exists(str(tmpdir))


def test_in_dir_from_config_dir(tmpdir):
    """config.in_dir() finds configs config dir."""

    cli.startup(str(tmpdir))
    tmpdir.join("myconfig.yaml").write("")
    tmpdir.join("myconfig.json").write("")
    configs_found = config.in_dir(str(tmpdir))

    assert len(configs_found) == 2


def test_ignore_non_configs_from_current_dir(tmpdir):
    """cli.in_dir() ignore non-config from config dir."""

    cli.startup(str(tmpdir))

    tmpdir.join("myconfig.psd").write("")
    tmpdir.join("watmyconfig.json").write("")
    configs_found = config.in_dir(str(tmpdir))
    assert len(configs_found) == 1


def test_get_configs_cwd(tmpdir):
    """config.in_cwd() find config in shell current working directory."""

    confdir = tmpdir.mkdir("tmuxpconf2")
    with confdir.as_cwd():
        config1 = open('.tmuxp.json', 'w+b')
        config1.close()

        configs_found = config.in_cwd()
        assert len(configs_found) == 1
        assert '.tmuxp.json' in configs_found


@pytest.mark.parametrize(
    'path,expect',
    [
        ('.', False),
        ('./', False),
        ('', False),
        ('.tmuxp.yaml', False),
        ('../.tmuxp.yaml', False),
        ('../', False),
        ('/hello/world', False),
        ('~/.tmuxp/hey', False),
        ('~/work/c/tmux/', False),
        ('~/work/c/tmux/.tmuxp.yaml', False),
        ('myproject', True),
    ],
)
def test_is_pure_name(path, expect):
    assert is_pure_name(path) == expect


"""
    scans for .tmuxp.{yaml,yml,json} in directory, returns first result
    log warning if multiple found:

    - current directory: ., ./, noarg
    - relative to cwd directory: ../, ./hello/, hello/, ./hello/
    - absolute directory: /path/to/dir, /path/to/dir/, ~/
    - no path, no ext, config_dir: projectname, tmuxp

    load file directly -

    - no directory (cwd): .tmuxp.yaml
    - relative to cwd: ../.tmuxp.yaml, ./hello/.tmuxp.yaml
    - absolute path: /path/to/file.yaml, ~/path/to/file/.tmuxp.yaml

    Any case where file is not found should return error.
"""


@pytest.fixture
def homedir(tmpdir):
    return tmpdir.join('home').mkdir()


@pytest.fixture
def configdir(homedir):
    return homedir.join('.tmuxp').mkdir()


@pytest.fixture
def projectdir(homedir):
    return homedir.join('work').join('project')


def test_tmuxp_configdir_env_var(tmpdir, monkeypatch):
    monkeypatch.setenv('TMUXP_CONFIGDIR', str(tmpdir))

    assert get_config_dir() == tmpdir


def test_tmuxp_configdir_xdg_config_dir(tmpdir, monkeypatch):
    monkeypatch.setenv('XDG_CONFIG_HOME', str(tmpdir))
    tmux_dir = tmpdir.mkdir("tmuxp")

    assert get_config_dir() == str(tmux_dir)


def test_resolve_dot(tmpdir, homedir, configdir, projectdir, monkeypatch):
    monkeypatch.setenv('HOME', str(homedir))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(homedir.join('.config')))

    projectdir.join('.tmuxp.yaml').ensure()
    user_config_name = 'myconfig'
    user_config = configdir.join('%s.yaml' % user_config_name).ensure()

    project_config = str(projectdir.join('.tmuxp.yaml'))

    with projectdir.as_cwd():
        expect = project_config
        assert scan_config('.') == expect
        assert scan_config('./') == expect
        assert scan_config('') == expect
        assert scan_config('../project') == expect
        assert scan_config('../project/') == expect
        assert scan_config('.tmuxp.yaml') == expect
        assert scan_config('../../.tmuxp/%s.yaml' % user_config_name) == str(
            user_config
        )
        assert scan_config('myconfig') == str(user_config)
        assert scan_config('~/.tmuxp/myconfig.yaml') == str(user_config)

        with pytest.raises(Exception):
            scan_config('.tmuxp.json')
        with pytest.raises(Exception):
            scan_config('.tmuxp.ini')
        with pytest.raises(Exception):
            scan_config('../')
        with pytest.raises(Exception):
            scan_config('mooooooo')

    with homedir.as_cwd():
        expect = project_config
        assert scan_config('work/project') == expect
        assert scan_config('work/project/') == expect
        assert scan_config('./work/project') == expect
        assert scan_config('./work/project/') == expect
        assert scan_config('.tmuxp/%s.yaml' % user_config_name) == str(user_config)
        assert scan_config('./.tmuxp/%s.yaml' % user_config_name) == str(user_config)
        assert scan_config('myconfig') == str(user_config)
        assert scan_config('~/.tmuxp/myconfig.yaml') == str(user_config)

        with pytest.raises(Exception):
            scan_config('')
        with pytest.raises(Exception):
            scan_config('.')
        with pytest.raises(Exception):
            scan_config('.tmuxp.yaml')
        with pytest.raises(Exception):
            scan_config('../')
        with pytest.raises(Exception):
            scan_config('mooooooo')

    with configdir.as_cwd():
        expect = project_config
        assert scan_config('../work/project') == expect
        assert scan_config('../../home/work/project') == expect
        assert scan_config('../work/project/') == expect
        assert scan_config('%s.yaml' % user_config_name) == str(user_config)
        assert scan_config('./%s.yaml' % user_config_name) == str(user_config)
        assert scan_config('myconfig') == str(user_config)
        assert scan_config('~/.tmuxp/myconfig.yaml') == str(user_config)

        with pytest.raises(Exception):
            scan_config('')
        with pytest.raises(Exception):
            scan_config('.')
        with pytest.raises(Exception):
            scan_config('.tmuxp.yaml')
        with pytest.raises(Exception):
            scan_config('../')
        with pytest.raises(Exception):
            scan_config('mooooooo')

    with tmpdir.as_cwd():
        expect = project_config
        assert scan_config('home/work/project') == expect
        assert scan_config('./home/work/project/') == expect
        assert scan_config('home/.tmuxp/%s.yaml' % user_config_name) == str(user_config)
        assert scan_config('./home/.tmuxp/%s.yaml' % user_config_name) == str(
            user_config
        )
        assert scan_config('myconfig') == str(user_config)
        assert scan_config('~/.tmuxp/myconfig.yaml') == str(user_config)

        with pytest.raises(Exception):
            scan_config('')
        with pytest.raises(Exception):
            scan_config('.')
        with pytest.raises(Exception):
            scan_config('.tmuxp.yaml')
        with pytest.raises(Exception):
            scan_config('../')
        with pytest.raises(Exception):
            scan_config('mooooooo')


def test_scan_config_arg(homedir, configdir, projectdir, monkeypatch):
    runner = CliRunner()

    @click.command()
    @click.argument('config', type=cli.ConfigPath(exists=True), nargs=-1)
    def config_cmd(config):
        click.echo(config)

    monkeypatch.setenv('HOME', str(homedir))
    projectdir.join('.tmuxp.yaml').ensure()
    user_config_name = 'myconfig'
    user_config = configdir.join('%s.yaml' % user_config_name).ensure()

    project_config = str(projectdir.join('.tmuxp.yaml'))

    def check_cmd(config_arg):
        return runner.invoke(config_cmd, [config_arg]).output

    with projectdir.as_cwd():
        expect = project_config
        assert expect in check_cmd('.')
        assert expect in check_cmd('./')
        assert expect in check_cmd('')
        assert expect in check_cmd('../project')
        assert expect in check_cmd('../project/')
        assert expect in check_cmd('.tmuxp.yaml')
        assert str(user_config) in check_cmd('../../.tmuxp/%s.yaml' % user_config_name)
        assert user_config.purebasename in check_cmd('myconfig')
        assert str(user_config) in check_cmd('~/.tmuxp/myconfig.yaml')

        assert 'file not found' in check_cmd('.tmuxp.json')
        assert 'file not found' in check_cmd('.tmuxp.ini')
        assert 'No tmuxp files found' in check_cmd('../')
        assert 'config not found in config dir' in check_cmd('moo')


def test_load_workspace(server, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv('TMUX', raising=False)
    session_file = curjoin("workspacebuilder/two_pane.yaml")

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == 'sampleconfig'


def test_load_workspace_named_session(server, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv('TMUX', raising=False)
    session_file = curjoin("workspacebuilder/two_pane.yaml")

    # open it detached
    session = load_workspace(
        session_file,
        socket_name=server.socket_name,
        new_session_name='tmuxp-new',
        detached=True,
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == 'tmuxp-new'


@pytest.mark.skipif(
    has_lt_version('2.1'), reason='exact session name matches only tmux >= 2.1'
)
def test_load_workspace_name_match_regression_252(tmpdir, server, monkeypatch):
    monkeypatch.delenv('TMUX', raising=False)
    session_file = curjoin("workspacebuilder/two_pane.yaml")

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == 'sampleconfig'

    projfile = tmpdir.join('simple.yaml')

    projfile.write(
        """
session_name: sampleconfi
start_directory: './'
windows:
- panes:
    - echo 'hey'"""
    )

    # open it detached
    session = load_workspace(
        projfile.strpath, socket_name=server.socket_name, detached=True
    )
    assert session.name == 'sampleconfi'


def test_load_symlinked_workspace(server, tmpdir, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv('TMUX', raising=False)

    realtemp = tmpdir.mkdir('myrealtemp')
    linktemp = tmpdir.join('symlinktemp')
    linktemp.mksymlinkto(realtemp)
    projfile = linktemp.join('simple.yaml')

    projfile.write(
        """
session_name: samplesimple
start_directory: './'
windows:
- panes:
    - echo 'hey'"""
    )

    # open it detached
    session = load_workspace(
        projfile.strpath, socket_name=server.socket_name, detached=True
    )
    pane = session.attached_window.attached_pane

    assert isinstance(session, libtmux.Session)
    assert session.name == 'samplesimple'
    assert pane.current_path == realtemp.strpath


def test_regression_00132_session_name_with_dots(tmpdir, server, session):
    yaml_config = curjoin("workspacebuilder/regression_00132_dots.yaml")
    cli_args = [yaml_config]
    inputs = []
    runner = CliRunner()
    result = runner.invoke(
        cli.command_load, cli_args, input=''.join(inputs), standalone_mode=False
    )
    assert result.exception
    assert isinstance(result.exception, libtmux.exc.BadSessionName)


@pytest.mark.parametrize("cli_args", [(['load', '.']), (['load', '.tmuxp.yaml'])])
def test_load_zsh_autotitle_warning(cli_args, tmpdir, monkeypatch):
    # create dummy tmuxp yaml so we don't get yelled at
    tmpdir.join('.tmuxp.yaml').ensure()
    tmpdir.join('.oh-my-zsh').ensure(dir=True)
    monkeypatch.setenv('HOME', str(tmpdir))

    with tmpdir.as_cwd():
        runner = CliRunner()

        monkeypatch.delenv('DISABLE_AUTO_TITLE', raising=False)
        monkeypatch.setenv('SHELL', 'zsh')
        result = runner.invoke(cli.cli, cli_args)
        assert 'Please set' in result.output

        monkeypatch.setenv('DISABLE_AUTO_TITLE', 'false')
        result = runner.invoke(cli.cli, cli_args)
        assert 'Please set' in result.output

        monkeypatch.setenv('DISABLE_AUTO_TITLE', 'true')
        result = runner.invoke(cli.cli, cli_args)
        assert 'Please set' not in result.output

        monkeypatch.delenv('DISABLE_AUTO_TITLE', raising=False)
        monkeypatch.setenv('SHELL', 'sh')
        result = runner.invoke(cli.cli, cli_args)
        assert 'Please set' not in result.output


@pytest.mark.parametrize(
    "cli_args",
    [
        (['load', '.', '--log-file', 'log.txt']),
    ],
)
def test_load_log_file(cli_args, tmpdir, monkeypatch):
    # create dummy tmuxp yaml that breaks to prevent actually loading tmux
    tmpdir.join('.tmuxp.yaml').write(
        """
session_name: hello
        """
    )
    tmpdir.join('.oh-my-zsh').ensure(dir=True)
    monkeypatch.setenv('HOME', str(tmpdir))

    with tmpdir.as_cwd():
        print('tmpdir: {0}'.format(tmpdir))
        runner = CliRunner()

        # If autoconfirm (-y) no need to prompt y
        input_args = 'y\ny\n' if '-y' not in cli_args else ''

        runner.invoke(cli.cli, cli_args, input=input_args)
        assert 'Loading' in tmpdir.join('log.txt').open().read()


@pytest.mark.parametrize("cli_cmd", ['shell', ('shell', '--pdb')])
@pytest.mark.parametrize(
    "cli_args,inputs,env,expected_output",
    [
        (
            ['-L{SOCKET_NAME}', '-c', 'print(str(server.socket_name))'],
            [],
            {},
            '{SERVER_SOCKET_NAME}',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '{SESSION_NAME}',
                '-c',
                'print(session.name)',
            ],
            [],
            {},
            '{SESSION_NAME}',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '{SESSION_NAME}',
                '{WINDOW_NAME}',
                '-c',
                'print(server.has_session(session.name))',
            ],
            [],
            {},
            'True',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '{SESSION_NAME}',
                '{WINDOW_NAME}',
                '-c',
                'print(window.name)',
            ],
            [],
            {},
            '{WINDOW_NAME}',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '{SESSION_NAME}',
                '{WINDOW_NAME}',
                '-c',
                'print(pane.id)',
            ],
            [],
            {},
            '{PANE_ID}',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '-c',
                'print(pane.id)',
            ],
            [],
            {'TMUX_PANE': '{PANE_ID}'},
            '{PANE_ID}',
        ),
    ],
)
def test_shell(
    cli_cmd,
    cli_args,
    inputs,
    expected_output,
    env,
    tmpdir,
    monkeypatch,
    server,
    session,
):
    monkeypatch.setenv('HOME', str(tmpdir))
    window_name = 'my_window'
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=window_name,
        PANE_ID=window.attached_pane.id,
        SERVER_SOCKET_NAME=server.socket_name,
    )

    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    with tmpdir.as_cwd():
        runner = CliRunner()

        result = runner.invoke(
            cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
        )
        assert expected_output.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_cmd",
    [
        'shell',
        ('shell', '--pdb'),
    ],
)
@pytest.mark.parametrize(
    "cli_args,inputs,env,template_ctx,exception,message",
    [
        (
            ['-LDoesNotExist', '-c', 'print(str(server.socket_name))'],
            [],
            {},
            {},
            LibTmuxException,
            r'.*DoesNotExist.*',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                'nonexistant_session',
                '-c',
                'print(str(server.socket_name))',
            ],
            [],
            {},
            {'session_name': 'nonexistant_session'},
            exc.TmuxpException,
            'Session not found: nonexistant_session',
        ),
        (
            [
                '-L{SOCKET_NAME}',
                '{SESSION_NAME}',
                'nonexistant_window',
                '-c',
                'print(str(server.socket_name))',
            ],
            [],
            {},
            {'window_name': 'nonexistant_window'},
            exc.TmuxpException,
            'Window not found: {WINDOW_NAME}',
        ),
    ],
)
def test_shell_target_missing(
    cli_cmd,
    cli_args,
    inputs,
    env,
    template_ctx,
    exception,
    message,
    tmpdir,
    monkeypatch,
    socket_name,
    server,
    session,
):
    monkeypatch.setenv('HOME', str(tmpdir))
    window_name = 'my_window'
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=template_ctx.get('window_name', window_name),
        PANE_ID=template_ctx.get('pane_id'),
        SERVER_SOCKET_NAME=server.socket_name,
    )
    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    with tmpdir.as_cwd():
        runner = CliRunner()

        if exception is not None:
            with pytest.raises(exception, match=message.format(**template_ctx)):
                result = runner.invoke(
                    cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
                )
        else:
            result = runner.invoke(
                cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
            )
            assert message.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_cmd",
    [
        # 'shell',
        # ('shell', '--pdb'),
        ('shell', '--code'),
        # ('shell', '--bpython'),
        # ('shell', '--ptipython'),
        # ('shell', '--ptpython'),
        # ('shell', '--ipython'),
    ],
)
@pytest.mark.parametrize(
    "cli_args,inputs,env,message",
    [
        (
            [
                '-L{SOCKET_NAME}',
            ],
            [],
            {},
            '(InteractiveConsole)',
        ),
        (
            [
                '-L{SOCKET_NAME}',
            ],
            [],
            {'PANE_ID': '{PANE_ID}'},
            '(InteractiveConsole)',
        ),
    ],
)
def test_shell_plus(
    cli_cmd,
    cli_args,
    inputs,
    env,
    message,
    tmpdir,
    monkeypatch,
    server,
    session,
):
    monkeypatch.setenv('HOME', str(tmpdir))
    window_name = 'my_window'
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=window_name,
        PANE_ID=window.attached_pane.id,
        SERVER_SOCKET_NAME=server.socket_name,
    )

    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    with tmpdir.as_cwd():
        runner = CliRunner()

        result = runner.invoke(
            cli.cli, cli_args, input=''.join(inputs), catch_exceptions=True
        )
        assert message.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_args",
    [
        (['convert', '.']),
        (['convert', '.tmuxp.yaml']),
        (['convert', '.tmuxp.yaml', '-y']),
    ],
)
def test_convert(cli_args, tmpdir, monkeypatch):
    # create dummy tmuxp yaml so we don't get yelled at
    tmpdir.join('.tmuxp.yaml').write(
        """
session_name: hello
    """
    )
    tmpdir.join('.oh-my-zsh').ensure(dir=True)
    monkeypatch.setenv('HOME', str(tmpdir))

    with tmpdir.as_cwd():
        runner = CliRunner()

        # If autoconfirm (-y) no need to prompt y
        input_args = 'y\ny\n' if '-y' not in cli_args else ''

        runner.invoke(cli.cli, cli_args, input=input_args)
        assert tmpdir.join('.tmuxp.json').check()
        assert tmpdir.join('.tmuxp.json').open().read() == json.dumps(
            {'session_name': 'hello'}, indent=2
        )


@pytest.mark.parametrize("cli_args", [(['import'])])
def test_import(cli_args, monkeypatch):
    runner = CliRunner()

    result = runner.invoke(cli.cli, cli_args)
    assert 'tmuxinator' in result.output
    assert 'teamocil' in result.output


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ['import', 'teamocil', './.teamocil/config.yaml'],
            ['\n', 'y\n', './la.yaml\n', 'y\n'],
        ),
        (
            ['import', 'teamocil', './.teamocil/config.yaml'],
            ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n'],
        ),
        (
            ['import', 'teamocil', 'config'],
            ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n'],
        ),
    ],
)
def test_import_teamocil(cli_args, inputs, tmpdir, monkeypatch):
    teamocil_config = loadfixture('config_teamocil/test4.yaml')
    teamocil_dir = tmpdir.join('.teamocil').mkdir()
    teamocil_dir.join('config.yaml').write(teamocil_config)
    tmpdir.join('exists.yaml').ensure()
    monkeypatch.setenv('HOME', str(tmpdir))

    with tmpdir.as_cwd():
        runner = CliRunner()
        runner.invoke(cli.cli, cli_args, input=''.join(inputs))
        assert tmpdir.join('la.yaml').check()


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ['import', 'tmuxinator', './.tmuxinator/config.yaml'],
            ['\n', 'y\n', './la.yaml\n', 'y\n'],
        ),
        (
            ['import', 'tmuxinator', './.tmuxinator/config.yaml'],
            ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n'],
        ),
        (
            ['import', 'tmuxinator', 'config'],
            ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n'],
        ),
    ],
)
def test_import_tmuxinator(cli_args, inputs, tmpdir, monkeypatch):
    tmuxinator_config = loadfixture('config_tmuxinator/test3.yaml')
    tmuxinator_dir = tmpdir.join('.tmuxinator').mkdir()
    tmuxinator_dir.join('config.yaml').write(tmuxinator_config)
    tmpdir.join('exists.yaml').ensure()
    monkeypatch.setenv('HOME', str(tmpdir))

    with tmpdir.as_cwd():
        runner = CliRunner()
        out = runner.invoke(cli.cli, cli_args, input=''.join(inputs))
        print(out.output)
        assert tmpdir.join('la.yaml').check()


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (['freeze', 'mysession'], ['\n', 'y\n', './la.yaml\n', 'y\n']),
        (  # Exists
            ['freeze', 'mysession'],
            ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n'],
        ),
        (  # Imply current session if not entered
            ['freeze'],
            ['\n', 'y\n', './la.yaml\n', 'y\n'],
        ),
        (['freeze'], ['\n', 'y\n', './exists.yaml\n', './la.yaml\n', 'y\n']),  # Exists
        (  # Create a new one
            ['freeze', 'mysession', '--force'],
            ['\n', 'y\n', './la.yaml\n', 'y\n'],
        ),
        (  # Imply current session if not entered
            ['freeze', '--force'],
            ['\n', 'y\n', './la.yaml\n', 'y\n'],
        ),
    ],
)
def test_freeze(server, cli_args, inputs, tmpdir, monkeypatch):
    monkeypatch.setenv('HOME', str(tmpdir))
    tmpdir.join('exists.yaml').ensure()

    server.new_session(session_name='mysession')

    with tmpdir.as_cwd():
        runner = CliRunner()
        # Use tmux server (socket name) used in the test
        cli_args = cli_args + ['-L', server.socket_name]
        out = runner.invoke(cli.cli, cli_args, input=''.join(inputs))
        print(out.output)
        assert tmpdir.join('la.yaml').check()


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (  # Overwrite
            ['freeze', 'mysession', '--force'],
            ['\n', 'y\n', './exists.yaml\n', 'y\n'],
        ),
        (  # Imply current session if not entered
            ['freeze', '--force'],
            ['\n', 'y\n', './exists.yaml\n', 'y\n'],
        ),
    ],
)
def test_freeze_overwrite(server, cli_args, inputs, tmpdir, monkeypatch):
    monkeypatch.setenv('HOME', str(tmpdir))
    tmpdir.join('exists.yaml').ensure()

    server.new_session(session_name='mysession')

    with tmpdir.as_cwd():
        runner = CliRunner()
        # Use tmux server (socket name) used in the test
        cli_args = cli_args + ['-L', server.socket_name]
        out = runner.invoke(cli.cli, cli_args, input=''.join(inputs))
        print(out.output)
        assert tmpdir.join('exists.yaml').check()


def test_get_abs_path(tmpdir):
    expect = str(tmpdir)
    with tmpdir.as_cwd():
        cli.get_abs_path('../') == os.path.dirname(expect)
        cli.get_abs_path('.') == expect
        cli.get_abs_path('./') == expect
        cli.get_abs_path(expect) == expect


def test_get_tmuxinator_dir(monkeypatch):
    assert cli.get_tmuxinator_dir() == os.path.expanduser('~/.tmuxinator/')

    monkeypatch.setenv('HOME', '/moo')
    assert cli.get_tmuxinator_dir() == '/moo/.tmuxinator/'
    assert cli.get_tmuxinator_dir() == os.path.expanduser('~/.tmuxinator/')


def test_get_cwd(tmpdir):
    assert cli.get_cwd() == os.getcwd()

    with tmpdir.as_cwd():
        assert cli.get_cwd() == str(tmpdir)
        assert cli.get_cwd() == os.getcwd()


def test_get_teamocil_dir(monkeypatch):
    assert cli.get_teamocil_dir() == os.path.expanduser('~/.teamocil/')

    monkeypatch.setenv('HOME', '/moo')
    assert cli.get_teamocil_dir() == '/moo/.teamocil/'
    assert cli.get_teamocil_dir() == os.path.expanduser('~/.teamocil/')


def test_validate_choices():
    validate = cli._validate_choices(['choice1', 'choice2'])

    assert validate('choice1')
    assert validate('choice2')

    with pytest.raises(click.BadParameter):
        assert validate('choice3')


def test_pass_config_dir_ClickPath(tmpdir):
    configdir = tmpdir.join('myconfigdir')
    configdir.mkdir()
    user_config_name = 'myconfig'
    user_config = configdir.join('%s.yaml' % user_config_name).ensure()

    expect = str(configdir.join('myconfig.yaml'))

    runner = CliRunner()

    @click.command()
    @click.argument(
        'config',
        type=cli.ConfigPath(exists=True, config_dir=(str(configdir))),
        nargs=-1,
    )
    def config_cmd(config):
        click.echo(config)

    def check_cmd(config_arg):
        return runner.invoke(config_cmd, [config_arg]).output

    with configdir.as_cwd():
        assert expect in check_cmd('myconfig')
        assert expect in check_cmd('myconfig.yaml')
        assert expect in check_cmd('./myconfig.yaml')
        assert str(user_config) in check_cmd(str(configdir.join('myconfig.yaml')))

        assert 'file not found' in check_cmd('.tmuxp.json')


def test_ls_cli(monkeypatch, tmpdir):
    monkeypatch.setenv("HOME", str(tmpdir))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir.join('.config')))

    filenames = [
        '.git/',
        '.gitignore/',
        'session_1.yaml',
        'session_2.yaml',
        'session_3.json',
        'session_4.txt',
    ]

    # should ignore:
    # - directories should be ignored
    # - extensions not covered in VALID_CONFIG_DIR_FILE_EXTENSIONS
    ignored_filenames = ['.git/', '.gitignore/', 'session_4.txt']
    stems = [os.path.splitext(f)[0] for f in filenames if f not in ignored_filenames]

    for filename in filenames:
        location = tmpdir.join('.tmuxp/{}'.format(filename))
        if filename.endswith('/'):
            location.ensure_dir()
        else:
            location.ensure()

    runner = CliRunner()
    cli_output = runner.invoke(command_ls).output
    assert cli_output == '\n'.join(stems) + '\n'


def test_load_plugins(monkeypatch_plugin_test_packages):
    from tmuxp_test_plugin_bwb.plugin import PluginBeforeWorkspaceBuilder

    plugins_config = loadfixture("workspacebuilder/plugin_bwb.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(plugins_config).get()
    sconfig = config.expand(sconfig)

    plugins = load_plugins(sconfig)

    assert len(plugins) == 1

    test_plugin_class_types = [
        PluginBeforeWorkspaceBuilder().__class__,
    ]
    for plugin in plugins:
        assert plugin.__class__ in test_plugin_class_types


@pytest.mark.skip('Not sure how to clean up the tmux session this makes')
@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ['load', 'tests/fixtures/workspacebuilder/plugin_versions_fail.yaml'],
            ['y\n'],
        )
    ],
)
def test_load_plugins_version_fail_skip(
    monkeypatch_plugin_test_packages, cli_args, inputs
):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args, input=''.join(inputs))
    assert '[Loading]' in results.output


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ['load', 'tests/fixtures/workspacebuilder/plugin_versions_fail.yaml'],
            ['n\n'],
        )
    ],
)
def test_load_plugins_version_fail_no_skip(
    monkeypatch_plugin_test_packages, cli_args, inputs
):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args, input=''.join(inputs))
    assert '[Not Skipping]' in results.output


@pytest.mark.parametrize(
    "cli_args", [(['load', 'tests/fixtures/workspacebuilder/plugin_missing_fail.yaml'])]
)
def test_load_plugins_plugin_missing(monkeypatch_plugin_test_packages, cli_args):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args)
    assert '[Plugin Error]' in results.output


def test_plugin_system_before_script(
    monkeypatch_plugin_test_packages, server, monkeypatch
):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv('TMUX', raising=False)
    session_file = curjoin("workspacebuilder/plugin_bs.yaml")

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == 'plugin_test_bs'


def test_reattach_plugins(monkeypatch_plugin_test_packages, server):
    config_plugins = loadfixture("workspacebuilder/plugin_r.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    # open it detached
    builder = WorkspaceBuilder(
        sconf=sconfig, plugins=load_plugins(sconfig), server=server
    )
    builder.build()

    try:
        _reattach(builder)
    except libtmux.exc.LibTmuxException:
        pass

    proc = builder.session.cmd('display-message', '-p', "'#S'")

    assert proc.stdout[0] == "'plugin_test_r'"


def test_load_attached(server, monkeypatch):
    # Load a session and attach from outside tmux
    monkeypatch.delenv('TMUX', raising=False)

    attach_session_mock = MagicMock()
    attach_session_mock.return_value.stderr = None

    monkeypatch.setattr("libtmux.session.Session.attach_session", attach_session_mock)

    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, False)

    assert builder.session.attach_session.call_count == 1


def test_load_attached_detached(server, monkeypatch):
    # Load a session but don't attach
    monkeypatch.delenv('TMUX', raising=False)

    attach_session_mock = MagicMock()
    attach_session_mock.return_value.stderr = None

    monkeypatch.setattr("libtmux.session.Session.attach_session", attach_session_mock)

    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, True)

    assert builder.session.attach_session.call_count == 0


def test_load_attached_within_tmux(server, monkeypatch):
    # Load a session and attach from within tmux
    monkeypatch.setenv('TMUX', "/tmp/tmux-1234/default,123,0")

    switch_client_mock = MagicMock()
    switch_client_mock.return_value.stderr = None

    monkeypatch.setattr("libtmux.session.Session.switch_client", switch_client_mock)

    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, False)

    assert builder.session.switch_client.call_count == 1


def test_load_attached_within_tmux_detached(server, monkeypatch):
    # Load a session and attach from within tmux
    monkeypatch.setenv('TMUX', "/tmp/tmux-1234/default,123,0")

    switch_client_mock = MagicMock()
    switch_client_mock.return_value.stderr = None

    monkeypatch.setattr("libtmux.session.Session.switch_client", switch_client_mock)

    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, True)

    assert builder.session.switch_client.call_count == 1


def test_load_append_windows_to_current_session(server, monkeypatch):
    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.list_sessions()) == 1
    assert len(server._list_windows()) == 3

    # Assign an active pane to the session
    monkeypatch.setenv("TMUX_PANE", server._list_panes()[0]["pane_id"])

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    _load_append_windows_to_current_session(builder)

    assert len(server.list_sessions()) == 1
    assert len(server._list_windows()) == 6


def test_debug_info_cli(monkeypatch, tmpdir):
    monkeypatch.setenv('SHELL', '/bin/bash')

    runner = CliRunner()
    cli_output = runner.invoke(command_debug_info).output
    assert 'environment' in cli_output
    assert 'python version' in cli_output
    assert 'system PATH' in cli_output
    assert 'tmux version' in cli_output
    assert 'libtmux version' in cli_output
    assert 'tmuxp version' in cli_output
    assert 'tmux path' in cli_output
    assert 'tmuxp path' in cli_output
    assert 'shell' in cli_output
    assert 'tmux session' in cli_output
    assert 'tmux windows' in cli_output
    assert 'tmux panes' in cli_output
    assert 'tmux global options' in cli_output
    assert 'tmux window options' in cli_output



"""Create a tmux workspace from a configuration :py:obj:`dict`.

tmuxp.workspacebuilder
~~~~~~~~~~~~~~~~~~~~~~

"""
import logging

from libtmux.exc import TmuxSessionExists
from libtmux.pane import Pane
from libtmux.server import Server
from libtmux.session import Session
from libtmux.window import Window

from . import exc
from .util import get_current_pane, run_before_script

logger = logging.getLogger(__name__)


class WorkspaceBuilder(object):

    """
    Load workspace from session :py:obj:`dict`.

    Build tmux workspace from a configuration. Creates and names windows, sets
    options, splits windows into panes.

    The normal phase of loading is:

    1. :term:`kaptan` imports json/yaml/ini. ``.get()`` returns python
       :class:`dict`::

           import kaptan
           sconf = kaptan.Kaptan(handler='yaml')
           sconf = sconfig.import_config(self.yaml_config).get()

       or from config file with extension::

           import kaptan
           sconf = kaptan.Kaptan()
           sconf = sconfig.import_config('path/to/config.yaml').get()

       kaptan automatically detects the handler from filenames.

    2. :meth:`config.expand` sconf inline shorthand::

           from tmuxp import config
           sconf = config.expand(sconf)

    3. :meth:`config.trickle` passes down default values from session
       -> window -> pane if applicable::

           sconf = config.trickle(sconf)

    4. (You are here) We will create a :class:`libtmux.Session` (a real
       ``tmux(1)`` session) and iterate through the list of windows, and
       their panes, returning full :class:`libtmux.Window` and
       :class:`libtmux.Pane` objects each step of the way::

           workspace = WorkspaceBuilder(sconf=sconf)

    It handles the magic of cases where the user may want to start
    a session inside tmux (when `$TMUX` is in the env variables).
    """

    def __init__(self, sconf, plugins=[], server=None):
        """
        Initialize workspace loading.

        Parameters
        ----------
        sconf : dict
            session config, includes a :py:obj:`list` of ``windows``.

        plugins : list
            plugins to be used for this session

        server : :class:`libtmux.Server`
            tmux server to build session in

        Notes
        -----
        TODO: Initialize :class:`libtmux.Session` from here, in
        ``self.session``.
        """

        if not sconf:
            raise exc.EmptyConfigException('session configuration is empty.')

        # config.validate_schema(sconf)

        if isinstance(server, Server):
            self.server = server
        else:
            self.server = None

        self.sconf = sconf

        self.plugins = plugins

    def session_exists(self, session_name=None):
        exists = self.server.has_session(session_name)
        if not exists:
            return exists

        self.session = self.server.find_where({'session_name': session_name})
        return True

    def build(self, session=None, append=False):
        """
        Build tmux workspace in session.

        Optionally accepts ``session`` to build with only session object.

        Without ``session``, it will use :class:`libmtux.Server` at
        ``self.server`` passed in on initialization to create a new Session
        object.

        Parameters
        ----------
        session : :class:`libtmux.Session`
            session to build workspace in
        append : bool
            append windows in current active session
        """

        if not session:
            if not self.server:
                raise exc.TmuxpException(
                    'WorkspaceBuilder.build requires server to be passed '
                    + 'on initialization, or pass in session object to here.'
                )

            if self.server.has_session(self.sconf['session_name']):
                self.session = self.server.find_where(
                    {'session_name': self.sconf['session_name']}
                )
                raise TmuxSessionExists(
                    'Session name %s is already running.' % self.sconf['session_name']
                )
            else:
                if 'start_directory' in self.sconf:
                    session = self.server.new_session(
                        session_name=self.sconf['session_name'],
                        start_directory=self.sconf['start_directory'],
                    )
                else:
                    session = self.server.new_session(
                        session_name=self.sconf['session_name']
                    )

            assert self.sconf['session_name'] == session.name
            assert len(self.sconf['session_name']) > 0

        self.session = session
        self.server = session.server

        self.server._list_sessions()
        assert self.server.has_session(session.name)
        assert session.id

        assert isinstance(session, Session)

        for plugin in self.plugins:
            plugin.before_workspace_builder(self.session)

        focus = None

        if 'before_script' in self.sconf:
            try:
                cwd = None

                # we want to run the before_script file cwd'd from the
                # session start directory, if it exists.
                if 'start_directory' in self.sconf:
                    cwd = self.sconf['start_directory']
                run_before_script(self.sconf['before_script'], cwd=cwd)
            except Exception as e:
                self.session.kill_session()
                raise e
        if 'options' in self.sconf:
            for option, value in self.sconf['options'].items():
                self.session.set_option(option, value)
        if 'global_options' in self.sconf:
            for option, value in self.sconf['global_options'].items():
                self.session.set_option(option, value, _global=True)
        if 'environment' in self.sconf:
            for option, value in self.sconf['environment'].items():
                self.session.set_environment(option, value)

        for w, wconf in self.iter_create_windows(session, append):
            assert isinstance(w, Window)

            for plugin in self.plugins:
                plugin.on_window_create(w)

            focus_pane = None
            for p, pconf in self.iter_create_panes(w, wconf):
                assert isinstance(p, Pane)
                p = p

                if 'layout' in wconf:
                    w.select_layout(wconf['layout'])

                if 'focus' in pconf and pconf['focus']:
                    focus_pane = p

            if 'focus' in wconf and wconf['focus']:
                focus = w

            self.config_after_window(w, wconf)

            for plugin in self.plugins:
                plugin.after_window_finished(w)

            if focus_pane:
                focus_pane.select_pane()

        if focus:
            focus.select_window()

    def iter_create_windows(self, session, append=False):
        """
        Return :class:`libtmux.Window` iterating through session config dict.

        Generator yielding :class:`libtmux.Window` by iterating through
        ``sconf['windows']``.

        Applies ``window_options`` to window.

        Parameters
        ----------
        session : :class:`libtmux.Session`
            session to create windows in
        append : bool
            append windows in current active session

        Returns
        -------
        tuple of (:class:`libtmux.Window`, ``wconf``)
            Newly created window, and the section from the tmuxp configuration
            that was used to create the window.
        """
        for i, wconf in enumerate(self.sconf['windows'], start=1):
            if 'window_name' not in wconf:
                window_name = None
            else:
                window_name = wconf['window_name']

            is_first_window_pass = self.first_window_pass(i, session, append)

            w1 = None
            if is_first_window_pass:  # if first window, use window 1
                w1 = session.attached_window
                w1.move_window(99)

            if 'start_directory' in wconf:
                sd = wconf['start_directory']
            else:
                sd = None

            if 'window_shell' in wconf:
                ws = wconf['window_shell']
            else:
                ws = None

            w = session.new_window(
                window_name=window_name,
                start_directory=sd,
                attach=False,  # do not move to the new window
                window_index=wconf.get('window_index', ''),
                window_shell=ws,
            )

            if is_first_window_pass:  # if first window, use window 1
                session.attached_window.kill_window()

            assert isinstance(w, Window)
            session.server._update_windows()
            if 'options' in wconf and isinstance(wconf['options'], dict):
                for key, val in wconf['options'].items():
                    w.set_window_option(key, val)

            if 'focus' in wconf and wconf['focus']:
                w.select_window()

            session.server._update_windows()

            yield w, wconf

    def iter_create_panes(self, w, wconf):
        """
        Return :class:`libtmux.Pane` iterating through window config dict.

        Run ``shell_command`` with ``$ tmux send-keys``.

        Parameters
        ----------
        w : :class:`libtmux.Window`
            window to create panes for
        wconf : dict
            config section for window

        Returns
        -------
        tuple of (:class:`libtmux.Pane`, ``pconf``)
            Newly created pane, and the section from the tmuxp configuration
            that was used to create the pane.
        """
        assert isinstance(w, Window)

        pane_base_index = int(w.show_window_option('pane-base-index', g=True))

        p = None

        for pindex, pconf in enumerate(wconf['panes'], start=pane_base_index):
            if pindex == int(pane_base_index):
                p = w.attached_pane
            else:

                def get_pane_start_directory():

                    if 'start_directory' in pconf:
                        return pconf['start_directory']
                    elif 'start_directory' in wconf:
                        return wconf['start_directory']
                    else:
                        return None

                p = w.split_window(
                    attach=True, start_directory=get_pane_start_directory(), target=p.id
                )

            assert isinstance(p, Pane)
            if 'layout' in wconf:
                w.select_layout(wconf['layout'])

            if 'suppress_history' in pconf:
                suppress = pconf['suppress_history']
            elif 'suppress_history' in wconf:
                suppress = wconf['suppress_history']
            else:
                suppress = True

            for cmd in pconf['shell_command']:
                p.send_keys(cmd, suppress_history=suppress)

            if 'focus' in pconf and pconf['focus']:
                w.select_pane(p['pane_id'])

            w.server._update_panes()

            yield p, pconf

    def config_after_window(self, w, wconf):
        """Actions to apply to window after window and pane finished.

        When building a tmux session, sometimes its easier to postpone things
        like setting options until after things are already structurally
        prepared.

        Parameters
        ----------
        w : :class:`libtmux.Window`
            window to create panes for
        wconf : dict
            config section for window
        """
        if 'options_after' in wconf and isinstance(wconf['options_after'], dict):
            for key, val in wconf['options_after'].items():
                w.set_window_option(key, val)

    def find_current_attached_session(self):
        current_active_pane = get_current_pane(self.server)

        if not current_active_pane:
            raise exc.TmuxpException("No session active.")

        return next(
            (
                s
                for s in self.server.list_sessions()
                if s["session_id"] == current_active_pane["session_id"]
            ),
            None,
        )

    def first_window_pass(self, i, session, append):
        return len(session.windows) == 1 and i == 1 and not append


def freeze(session):
    """
    Freeze live tmux session and Return session config :py:obj:`dict`.

    Parameters
    ----------
    session : :class:`libtmux.Session`
        session object

    Returns
    -------
    dict
        tmuxp compatible workspace config
    """
    sconf = {'session_name': session['session_name'], 'windows': []}

    for w in session.windows:
        wconf = {
            'options': w.show_window_options(),
            'window_name': w.name,
            'layout': w.layout,
            'panes': [],
        }
        if w.get('window_active', '0') == '1':
            wconf['focus'] = 'true'

        # If all panes have same path, set 'start_directory' instead
        # of using 'cd' shell commands.
        def pane_has_same_path(p):
            return w.panes[0].current_path == p.current_path

        if all(pane_has_same_path(p) for p in w.panes):
            wconf['start_directory'] = w.panes[0].current_path

        for p in w.panes:
            pconf = {'shell_command': []}

            if 'start_directory' not in wconf:
                pconf['shell_command'].append('cd ' + p.current_path)

            if p.get('pane_active', '0') == '1':
                pconf['focus'] = 'true'

            current_cmd = p.current_command

            def filter_interpretters_and_shells():
                return current_cmd.startswith('-') or any(
                    current_cmd.endswith(cmd) for cmd in ['python', 'ruby', 'node']
                )

            if filter_interpretters_and_shells():
                current_cmd = None

            if current_cmd:
                pconf['shell_command'].append(current_cmd)
            else:
                if not len(pconf['shell_command']):
                    pconf = 'pane'

            wconf['panes'].append(pconf)

        sconf['windows'].append(wconf)

    return sconf



"""Test for tmuxp workspacebuilder."""
import os

import pytest

import kaptan

from libtmux import Window
from libtmux.common import has_gte_version
from libtmux.test import retry, temp_session
from tmuxp import config, exc
from tmuxp.cli import load_plugins
from tmuxp.workspacebuilder import WorkspaceBuilder

from . import example_dir, fixtures_dir
from .fixtures._util import loadfixture


def test_split_windows(session):
    yaml_config = loadfixture("workspacebuilder/two_pane.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(s._windows) == window_count
    for w, wconf in builder.iter_create_windows(s):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout('tiled')  # fix glitch with pane size
            p = p
            assert len(s._windows) == window_count
        assert isinstance(w, Window)

        assert len(s._windows) == window_count
        window_count += 1


def test_split_windows_three_pane(session):
    yaml_config = loadfixture("workspacebuilder/three_pane.yaml")

    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(s._windows)  # current window count
    assert len(s._windows) == window_count
    for w, wconf in builder.iter_create_windows(s):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout('tiled')  # fix glitch with pane size
            p = p
            assert len(s._windows) == window_count
        assert isinstance(w, Window)

        assert len(s._windows) == window_count
        window_count += 1
        w.set_window_option('main-pane-height', 50)
        w.select_layout(wconf['layout'])


def test_focus_pane_index(session):
    yaml_config = loadfixture('workspacebuilder/focus_and_pane.yaml')

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    builder.build(session=session)

    assert session.attached_window.name == 'focused window'

    pane_base_index = int(
        session.attached_window.show_window_option('pane-base-index', g=True)
    )

    if not pane_base_index:
        pane_base_index = 0
    else:
        pane_base_index = int(pane_base_index)

    # get the pane index for each pane
    pane_base_indexes = []
    for pane in session.attached_window.panes:
        pane_base_indexes.append(int(pane.index))

    pane_indexes_should_be = [pane_base_index + x for x in range(0, 3)]
    assert pane_indexes_should_be == pane_base_indexes

    w = session.attached_window

    assert w.name != 'man'

    pane_path = '/usr'

    while retry():
        p = w.attached_pane
        p.server._update_panes()
        if p.current_path == pane_path:
            break

    assert p.current_path == pane_path

    proc = session.cmd('show-option', '-gv', 'base-index')
    base_index = int(proc.stdout[0])
    session.server._update_windows()

    window3 = session.find_where({'window_index': str(base_index + 2)})
    assert isinstance(window3, Window)

    p = None
    pane_path = '/'

    while retry():
        p = window3.attached_pane
        p.server._update_panes()
        if p.current_path == pane_path:
            break

    assert p.current_path == pane_path


@pytest.mark.skip(
    reason='''
Test needs to be rewritten, assertion not reliable across platforms
and CI. See https://github.com/tmux-python/tmuxp/issues/310.
    '''.strip()
)
def test_suppress_history(session):
    yaml_config = loadfixture("workspacebuilder/suppress_history.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    inHistoryWindow = session.find_where({'window_name': 'inHistory'})
    isMissingWindow = session.find_where({'window_name': 'isMissing'})

    def assertHistory(cmd, hist):
        return 'inHistory' in cmd and cmd.endswith(hist)

    def assertIsMissing(cmd, hist):
        return 'isMissing' in cmd and not cmd.endswith(hist)

    for w, window_name, assertCase in [
        (inHistoryWindow, 'inHistory', assertHistory),
        (isMissingWindow, 'isMissing', assertIsMissing),
    ]:
        assert w.name == window_name
        correct = False
        w.select_window()
        p = w.attached_pane
        p.select_pane()

        # Print the last-in-history command in the pane
        p.cmd('send-keys', ' fc -ln -1')
        p.cmd('send-keys', 'Enter')

        buffer_name = 'test'
        while retry():
            # from v0.7.4 libtmux session.cmd adds target -t self.id by default
            # show-buffer doesn't accept -t, use global cmd.

            # Get the contents of the pane
            p.cmd('capture-pane', '-b', buffer_name)

            captured_pane = session.server.cmd('show-buffer', '-b', buffer_name)
            session.server.cmd('delete-buffer', '-b', buffer_name)

            # Parse the sent and last-in-history commands
            sent_cmd = captured_pane.stdout[0].strip()
            history_cmd = captured_pane.stdout[-2].strip()

            if assertCase(sent_cmd, history_cmd):
                correct = True
                break
        assert correct, "Unknown sent command: [%s] in %s" % (sent_cmd, assertCase)


def test_session_options(session):
    yaml_config = loadfixture("workspacebuilder/session_options.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert "/bin/sh" in s.show_option('default-shell')
    assert "/bin/sh" in s.show_option('default-command')


def test_global_options(session):
    yaml_config = loadfixture("workspacebuilder/global_options.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert "top" in s.show_option('status-position', _global=True)
    assert 493 == s.show_option('repeat-time', _global=True)


def test_global_session_env_options(session, monkeypatch):
    visual_silence = 'on'
    monkeypatch.setenv(str('VISUAL_SILENCE'), str(visual_silence))
    repeat_time = 738
    monkeypatch.setenv(str('REPEAT_TIME'), str(repeat_time))
    main_pane_height = 8
    monkeypatch.setenv(str('MAIN_PANE_HEIGHT'), str(main_pane_height))

    yaml_config = loadfixture("workspacebuilder/env_var_options.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert visual_silence in s.show_option('visual-silence', _global=True)
    assert repeat_time == s.show_option('repeat-time')
    assert main_pane_height == s.attached_window.show_window_option('main-pane-height')


def test_window_options(session):
    yaml_config = loadfixture("workspacebuilder/window_options.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    if has_gte_version('2.3'):
        sconfig['windows'][0]['options']['pane-border-format'] = ' #P '

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(s._windows) == window_count
    for w, wconf in builder.iter_create_windows(s):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout('tiled')  # fix glitch with pane size
            p = p
            assert len(s._windows) == window_count
        assert isinstance(w, Window)
        assert w.show_window_option('main-pane-height') == 5
        if has_gte_version('2.3'):
            assert w.show_window_option('pane-border-format') == ' #P '

        assert len(s._windows) == window_count
        window_count += 1
        w.select_layout(wconf['layout'])


@pytest.mark.flaky(reruns=5)
def test_window_options_after(session):
    yaml_config = loadfixture("workspacebuilder/window_options_after.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    def assert_last_line(p, s):
        correct = False

        while retry():
            pane_out = p.cmd('capture-pane', '-p', '-J').stdout
            while not pane_out[-1].strip():  # delete trailing lines tmux 1.8
                pane_out.pop()
            if len(pane_out) > 1 and pane_out[-2].strip() == s:
                correct = True
                break

        # Print output for easier debugging if assertion fails
        if not correct:
            print('\n'.join(pane_out))

        return correct

    for i, pane in enumerate(session.attached_window.panes):
        assert assert_last_line(
            pane, str(i)
        ), "Initial command did not execute properly/" + str(i)
        pane.cmd('send-keys', 'Up')  # Will repeat echo
        pane.enter()  # in each iteration
        assert assert_last_line(
            pane, str(i)
        ), "Repeated command did not execute properly/" + str(i)

    session.cmd('send-keys', ' echo moo')
    session.cmd('send-keys', 'Enter')

    for pane in session.attached_window.panes:
        assert assert_last_line(
            pane, 'moo'
        ), "Synchronized command did not execute properly"


def test_window_shell(session):
    yaml_config = loadfixture("workspacebuilder/window_shell.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    for w, wconf in builder.iter_create_windows(s):
        if 'window_shell' in wconf:
            assert wconf['window_shell'] == str('top')

        while retry():
            session.server._update_windows()
            if w['window_name'] != 'top':
                break

        assert w.name != str('top')


def test_environment_variables(session):
    yaml_config = loadfixture("workspacebuilder/environment_vars.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session)

    assert session.show_environment('FOO') == 'BAR'
    assert session.show_environment('PATH') == '/tmp'


def test_automatic_rename_option(session):
    """With option automatic-rename: on."""
    yaml_config = loadfixture("workspacebuilder/window_automatic_rename.yaml")
    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(s._windows) == window_count
    for w, wconf in builder.iter_create_windows(s):

        for p in builder.iter_create_panes(w, wconf):
            w.select_layout('tiled')  # fix glitch with pane size
            p = p
            assert len(s._windows), window_count
        assert isinstance(w, Window)
        assert w.show_window_option('automatic-rename') == 'on'

        assert len(s._windows) == window_count

        window_count += 1
        w.select_layout(wconf['layout'])

    assert s.name != 'tmuxp'
    w = s.windows[0]

    while retry():
        session.server._update_windows()
        if w.name != 'sh':
            break

    assert w.name != 'sh'

    pane_base_index = w.show_window_option('pane-base-index', g=True)
    w.select_pane(pane_base_index)

    while retry():
        session.server._update_windows()
        if w.name == 'sh':
            break

    assert w.name == 'sh'

    w.select_pane('-D')

    while retry():
        session.server._update_windows()
        if w['window_name'] != 'sh':
            break

    assert w.name != 'sh'


def test_blank_pane_count(session):
    """:todo: Verify blank panes of various types build into workspaces."""
    yaml_config_file = os.path.join(example_dir, 'blank-panes.yaml')
    test_config = kaptan.Kaptan().import_config(yaml_config_file).get()
    test_config = config.expand(test_config)
    builder = WorkspaceBuilder(sconf=test_config)
    builder.build(session=session)

    assert session == builder.session

    window1 = session.find_where({'window_name': 'Blank pane test'})
    assert len(window1._panes) == 3

    window2 = session.find_where({'window_name': 'More blank panes'})
    assert len(window2._panes) == 3

    window3 = session.find_where({'window_name': 'Empty string (return)'})
    assert len(window3._panes) == 3

    window4 = session.find_where({'window_name': 'Blank with options'})
    assert len(window4._panes) == 2


def test_start_directory(session, tmpdir):
    yaml_config = loadfixture("workspacebuilder/start_directory.yaml")
    test_dir = str(tmpdir.mkdir('foo bar'))
    test_config = yaml_config.format(TMP_DIR=str(tmpdir), TEST_DIR=test_dir)

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(test_config).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session == builder.session
    dirs = ['/usr/bin', '/dev', test_dir, '/usr', '/usr']

    for path, window in zip(dirs, session.windows):
        for p in window.panes:
            while retry():
                p.server._update_panes()
                pane_path = p.current_path
                if pane_path is None:
                    pass
                elif path in pane_path or pane_path == path:
                    result = path == pane_path or path in pane_path
                    break

            # handle case with OS X adding /private/ to /tmp/ paths
            assert result


def test_start_directory_relative(session, tmpdir):
    """Same as above test, but with relative start directory, mimicking
    loading it from a location of project file. Like::

    $ tmuxp load ~/workspace/myproject/.tmuxp.yaml

    instead of::

    $ cd ~/workspace/myproject/.tmuxp.yaml
    $ tmuxp load .

    """
    yaml_config = loadfixture("workspacebuilder/start_directory_relative.yaml")

    test_dir = str(tmpdir.mkdir('foo bar'))
    config_dir = str(tmpdir.mkdir('testRelConfigDir'))
    test_config = yaml_config.format(TEST_DIR=test_dir)

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(test_config).get()
    # the second argument of os.getcwd() mimics the behavior
    # the CLI loader will do, but it passes in the config file's location.
    sconfig = config.expand(sconfig, config_dir)

    sconfig = config.trickle(sconfig)

    assert os.path.exists(config_dir)
    assert os.path.exists(test_dir)
    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session == builder.session

    dirs = ['/usr/bin', '/dev', test_dir, config_dir, config_dir]

    for path, window in zip(dirs, session.windows):
        for p in window.panes:
            while retry():
                p.server._update_panes()
                # Handle case where directories resolve to /private/ in OSX
                pane_path = p.current_path
                if pane_path is None:
                    pass
                elif path in pane_path or pane_path == path:
                    result = path == pane_path or path in pane_path
                    break

            assert result


def test_pane_order(session):
    """Pane ordering based on position in config and ``pane_index``.

    Regression test for https://github.com/tmux-python/tmuxp/issues/15.

    """

    yaml_config = loadfixture("workspacebuilder/pane_ordering.yaml").format(
        HOME=os.path.realpath(os.path.expanduser('~'))
    )

    # test order of `panes` (and pane_index) above aganist pane_dirs
    pane_paths = [
        '/usr/bin',
        '/usr',
        '/usr/sbin',
        os.path.realpath(os.path.expanduser('~')),
    ]

    s = session
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(s._windows) == window_count
    for w, wconf in builder.iter_create_windows(s):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout('tiled')  # fix glitch with pane size
            p = p
            assert len(s._windows) == window_count

        assert isinstance(w, Window)

        assert len(s._windows) == window_count
        window_count += 1

    for w in session.windows:
        pane_base_index = w.show_window_option('pane-base-index', g=True)
        for p_index, p in enumerate(w.list_panes(), start=pane_base_index):
            assert int(p_index) == int(p.index)

            # pane-base-index start at base-index, pane_paths always start
            # at 0 since python list.
            pane_path = pane_paths[p_index - pane_base_index]

            while retry():
                p.server._update_panes()
                if p.current_path == pane_path:
                    break

            assert p.current_path, pane_path


def test_window_index(session):
    yaml_config = loadfixture("workspacebuilder/window_index.yaml")
    proc = session.cmd('show-option', '-gv', 'base-index')
    base_index = int(proc.stdout[0])
    name_index_map = {'zero': 0 + base_index, 'one': 1 + base_index, 'five': 5}

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    for window, _ in builder.iter_create_windows(session):
        expected_index = name_index_map[window['window_name']]
        assert int(window['window_index']) == expected_index


def test_before_load_throw_error_if_retcode_error(server):
    config_script_fails = loadfixture("workspacebuilder/config_script_fails.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    yaml = config_script_fails.format(
        fixtures_dir=fixtures_dir,
        script_failed=os.path.join(fixtures_dir, 'script_failed.sh'),
    )

    sconfig = sconfig.import_config(yaml).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as sess:
        session_name = sess.name

        with pytest.raises(exc.BeforeLoadScriptError):
            builder.build(session=sess)

        result = server.has_session(session_name)
        assert not result, "Kills session if before_script exits with errcode"


def test_before_load_throw_error_if_file_not_exists(server):
    config_script_not_exists = loadfixture(
        "workspacebuilder/config_script_not_exists.yaml"
    )
    sconfig = kaptan.Kaptan(handler='yaml')
    yaml = config_script_not_exists.format(
        fixtures_dir=fixtures_dir,
        script_not_exists=os.path.join(fixtures_dir, 'script_not_exists.sh'),
    )
    sconfig = sconfig.import_config(yaml).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as sess:
        session_name = sess.name
        temp_session_exists = server.has_session(sess.name)
        assert temp_session_exists
        with pytest.raises((exc.BeforeLoadScriptNotExists, OSError)) as excinfo:
            builder.build(session=sess)
            excinfo.match(r'No such file or directory')
        result = server.has_session(session_name)
        assert not result, "Kills session if before_script doesn't exist"


def test_before_load_true_if_test_passes(server):
    config_script_completes = loadfixture(
        "workspacebuilder/config_script_completes.yaml"
    )
    assert os.path.exists(os.path.join(fixtures_dir, 'script_complete.sh'))
    sconfig = kaptan.Kaptan(handler='yaml')
    yaml = config_script_completes.format(
        fixtures_dir=fixtures_dir,
        script_complete=os.path.join(fixtures_dir, 'script_complete.sh'),
    )

    sconfig = sconfig.import_config(yaml).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as session:
        builder.build(session=session)


def test_before_load_true_if_test_passes_with_args(server):
    config_script_completes = loadfixture(
        "workspacebuilder/config_script_completes.yaml"
    )

    assert os.path.exists(os.path.join(fixtures_dir, 'script_complete.sh'))
    sconfig = kaptan.Kaptan(handler='yaml')
    yaml = config_script_completes.format(
        fixtures_dir=fixtures_dir,
        script_complete=os.path.join(fixtures_dir, 'script_complete.sh') + ' -v',
    )

    sconfig = sconfig.import_config(yaml).get()
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as session:
        builder.build(session=session)


def test_plugin_system_before_workspace_builder(
    monkeypatch_plugin_test_packages, session
):
    config_plugins = loadfixture("workspacebuilder/plugin_bwb.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd('display-message', '-p', "'#S'")
    assert proc.stdout[0] == "'plugin_test_bwb'"


def test_plugin_system_on_window_create(monkeypatch_plugin_test_packages, session):
    config_plugins = loadfixture("workspacebuilder/plugin_owc.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd('display-message', '-p', "'#W'")
    assert proc.stdout[0] == "'plugin_test_owc'"


def test_plugin_system_after_window_finished(monkeypatch_plugin_test_packages, session):
    config_plugins = loadfixture("workspacebuilder/plugin_awf.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd('display-message', '-p', "'#W'")
    assert proc.stdout[0] == "'plugin_test_awf'"


def test_plugin_system_on_window_create_multiple_windows(session):
    config_plugins = loadfixture("workspacebuilder/plugin_owc_multiple_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd('list-windows', '-F', "'#W'")
    assert "'plugin_test_owc_mw'" in proc.stdout
    assert "'plugin_test_owc_mw_2'" in proc.stdout


def test_plugin_system_after_window_finished_multiple_windows(
    monkeypatch_plugin_test_packages, session
):
    config_plugins = loadfixture("workspacebuilder/plugin_awf_multiple_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd('list-windows', '-F', "'#W'")
    assert "'plugin_test_awf_mw'" in proc.stdout
    assert "'plugin_test_awf_mw_2'" in proc.stdout


def test_plugin_system_multiple_plugins(monkeypatch_plugin_test_packages, session):
    config_plugins = loadfixture("workspacebuilder/plugin_multiple_plugins.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(config_plugins).get()
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    # Drop through to the before_script plugin hook
    proc = session.cmd('display-message', '-p', "'#S'")
    assert proc.stdout[0] == "'plugin_test_bwb'"

    # Drop through to the after_window_finished. This won't succeed
    # unless on_window_create succeeds because of how the test plugin
    # override methods are currently written
    proc = session.cmd('display-message', '-p', "'#W'")
    assert proc.stdout[0] == "'mp_test_awf'"


def test_load_configs_same_session(server):
    yaml_config = loadfixture("workspacebuilder/three_windows.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 1
    assert len(server.sessions[0]._windows) == 3

    yaml_config = loadfixture("workspacebuilder/two_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()
    assert len(server.sessions) == 2
    assert len(server.sessions[1]._windows) == 2

    yaml_config = loadfixture("workspacebuilder/two_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build(server.sessions[1], True)

    assert len(server.sessions) == 2
    assert len(server.sessions[1]._windows) == 4


def test_load_configs_separate_sessions(server):
    yaml_config = loadfixture("workspacebuilder/three_windows.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 1
    assert len(server.sessions[0]._windows) == 3

    yaml_config = loadfixture("workspacebuilder/two_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 2
    assert len(server.sessions[0]._windows) == 3
    assert len(server.sessions[1]._windows) == 2


def test_find_current_active_pane(server, monkeypatch):
    yaml_config = loadfixture("workspacebuilder/three_windows.yaml")
    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    yaml_config = loadfixture("workspacebuilder/two_windows.yaml")

    sconfig = kaptan.Kaptan(handler='yaml')
    sconfig = sconfig.import_config(yaml_config).get()

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.list_sessions()) == 2

    # Assign an active pane to the session
    second_session = server.list_sessions()[1]
    first_pane_on_second_session_id = second_session.list_windows()[0].list_panes()[0][
        "pane_id"
    ]
    monkeypatch.setenv("TMUX_PANE", first_pane_on_second_session_id)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    assert builder.find_current_attached_session() == second_session
